# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Unit tests for `grizzly.reduce.main`."""
from logging import getLogger
from pathlib import Path
from shutil import rmtree
from unittest.mock import Mock

import pytest
from pytest import raises

from ..common import TestCaseLoadFailure
from ..target import TargetLaunchError, TargetLaunchTimeout
from . import ReduceManager
from .args import ReduceArgs, ReduceFuzzManagerIDArgs, ReduceFuzzManagerIDQualityArgs
from .exceptions import GrizzlyReduceBaseException

LOG = getLogger(__name__)
pytestmark = pytest.mark.usefixtures("tmp_path_fm_config")


def test_args_01(capsys, tmp_path, mocker):
    """test args in common with grizzly.replay"""
    # pylint: disable=import-outside-toplevel
    from ..replay.test_args import test_replay_args_01 as real_test

    mocker.patch("grizzly.replay.test_args.ReplayArgs", new=ReduceArgs)
    real_test(capsys, mocker, tmp_path)


def test_args_02(tmp_path):
    """test parsing args specific to grizzly.reduce"""
    exe = tmp_path / "binary"
    exe.touch()
    inp = tmp_path / "input"
    inp.mkdir()
    (inp / "somefile").touch()

    # test valid strategy
    ReduceArgs().parse_args([str(exe), str(inp), "--strategy", "lines"])
    # test invalid strategy
    with raises(SystemExit):
        ReduceArgs().parse_args([str(exe), str(inp), "--strategy", "cosmic_radiation"])
    # test --logs must be dir
    logs_file = tmp_path / "logs1"
    logs_file.touch()
    with raises(SystemExit):
        ReduceArgs().parse_args([str(exe), str(inp), "--logs", str(logs_file)])
    logs_dir = tmp_path / "logs2"
    logs_dir.mkdir()
    ReduceArgs().parse_args([str(exe), str(inp), "--logs", str(logs_dir)])
    # test no-analysis
    ReduceArgs().parse_args(
        [str(exe), str(inp), "--no-analysis", "--repeat", "99", "--min-crashes", "99"]
    )
    # these should both log a warning that the args will be ignored due to analysis
    ReduceArgs().parse_args([str(exe), str(inp), "--repeat", "99"])
    ReduceArgs().parse_args([str(exe), str(inp), "--min-crashes", "99"])
    ReduceArgs().parse_args([str(exe), str(inp), "--report-period", "99"])
    with raises(SystemExit):
        ReduceArgs().parse_args([str(exe), str(inp), "--report-period", "0"])
    with raises(SystemExit):
        ReduceArgs().parse_args([str(exe), str(inp), "--report-period", "15"])


def test_args_03(tmp_path):
    """test ReduceFuzzManagerIDArgs"""
    exe = tmp_path / "binary"
    exe.touch()
    ReduceFuzzManagerIDArgs().parse_args([str(exe), "123"])


def test_args_04(capsys, tmp_path):
    """test ReduceFuzzManagerIDQualityArgs"""
    exe = tmp_path / "binary"
    exe.touch()
    with raises(SystemExit):
        ReduceFuzzManagerIDQualityArgs().parse_args(
            [str(exe), "123", "--quality", "-1"]
        )
    assert "error: '--quality' value cannot be negative" in capsys.readouterr()[-1]


@pytest.mark.parametrize(
    "patch_func, side_effect, return_value, kwargs, result",
    [
        (
            "grizzly.reduce.core.ReduceManager.run",
            TargetLaunchError("error", None),
            None,
            {},
            1,
        ),
        ("grizzly.reduce.core.ReduceManager.run", TargetLaunchTimeout, None, {}, 1),
        ("grizzly.reduce.core.load_plugin", KeyboardInterrupt, None, {}, 1),
        (
            "grizzly.reduce.core.load_plugin",
            GrizzlyReduceBaseException(""),
            None,
            {},
            1,
        ),
        (
            "grizzly.reduce.core.ReplayManager.load_testcases",
            TestCaseLoadFailure,
            None,
            {},
            1,
        ),
        ("grizzly.reduce.core.ReplayManager.load_testcases", None, [], {}, 1),
        (
            "grizzly.reduce.core.ReplayManager.load_testcases",
            None,
            [Mock(hang=False), Mock(hang=False)],
            {"no_harness": True},
            2,
        ),
    ],
)
def test_main_exit(mocker, patch_func, side_effect, return_value, kwargs, result):
    """test ReduceManager.main() failure cases"""
    mocker.patch(
        "grizzly.reduce.core.FuzzManagerReporter",
        autospec=True,
        QUAL_NO_TESTCASE=7,
        QUAL_REDUCER_ERROR=9,
    )
    mocker.patch("grizzly.reduce.core.load_plugin", autospec=True)
    mocker.patch("grizzly.reduce.core.Sapphire", autospec=True)
    # setup args
    args = mocker.Mock(
        ignore=["fake"],
        input="test",
        min_crashes=1,
        prefs=None,
        relaunch=1,
        repeat=1,
        sig=None,
        **kwargs
    )

    mocker.patch(patch_func, side_effect=side_effect, return_value=return_value)
    assert ReduceManager.main(args) == result


@pytest.mark.parametrize(
    "exc_type",
    [
        TargetLaunchError,
        TargetLaunchTimeout,
    ],
)
def test_main_launch_error(mocker, exc_type):
    mocker.patch("grizzly.reduce.core.FuzzManagerReporter", autospec=True)
    reporter = mocker.patch("grizzly.reduce.core.FilesystemReporter", autospec=True)
    mocker.patch("grizzly.reduce.core.load_plugin", autospec=True)
    mocker.patch("grizzly.reduce.core.ReplayManager.load_testcases")
    mocker.patch(
        "grizzly.reduce.core.ReplayManager.time_limits", return_value=(None, 10)
    )
    run = mocker.patch("grizzly.reduce.core.ReduceManager.run", autospec=True)
    mocker.patch("grizzly.reduce.core.Sapphire", autospec=True)
    # setup args
    args = mocker.Mock(
        ignore=["fake"],
        input="test",
        min_crashes=1,
        prefs=None,
        relaunch=1,
        repeat=1,
        sig=None,
    )

    exc_obj = exc_type(
        "error", mocker.Mock() if exc_type is TargetLaunchError else None
    )
    run.side_effect = exc_obj
    assert ReduceManager.main(args) == 4
    if exc_type is TargetLaunchError:
        assert reporter.return_value.submit.call_count == 1
        reported_testcases, report = reporter.return_value.submit.call_args[0]
        assert reported_testcases == []
        assert report is exc_obj.report
    else:
        assert reporter.return_value.submit.call_count == 0


@pytest.mark.parametrize("result", ["testprefs", "argprefs"])
def test_testcase_prefs(mocker, tmp_path, result):
    """test that prefs from testcase are used if --prefs not specified and --prefs
    overrides"""
    load_target = mocker.patch("grizzly.reduce.core.load_plugin")
    mocker.patch("grizzly.reduce.core.ReduceManager.run", autospec=True)
    mocker.patch("grizzly.reduce.core.Sapphire", autospec=True)
    rmtree_mock = mocker.patch("grizzly.reduce.core.rmtree", autospec=True)
    (tmp_path / "test.html").touch()
    (tmp_path / "prefs.js").write_text("testprefs")
    args = mocker.Mock(
        fuzzmanager=False,
        ignore=[],
        input=str(tmp_path / "test.html"),
        min_crashes=1,
        prefs=None,
        repeat=1,
        sig=None,
        test_index=None,
        time_limit=10,
        timeout=10,
    )
    if result == "argprefs":
        (tmp_path / "args.js").write_text("argprefs")
        args.prefs = str(tmp_path / "args.js")

    try:
        ReduceManager.main(args)
        assert Path(load_target.return_value.return_value.prefs).read_text() == result
    finally:
        for rm_args, rm_kwds in rmtree_mock.call_args_list:
            rmtree(*rm_args, **rm_kwds)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Grizzly reducer strategy definitions.

Each class defined here is an iterator yielding *potential* reductions. The caller
should evaluate each set of testcases, and keep the best one. The caller is responsible
for cleaning up all testcases that are yielded.

Constants:
    DEFAULT_STRATEGIES (list(str)): List of strategy names run by default if none are
                                    specified.
    STRATEGIES (dict{str: Strategy}): Mapping of available strategy names to
                                      implementing class.
"""
from abc import ABC, abstractmethod
from hashlib import sha512
from logging import DEBUG, getLogger
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from types import MappingProxyType

from pkg_resources import iter_entry_points

from ...common.utils import grz_tmp

LOG = getLogger(__name__)


DEFAULT_STRATEGIES = (
    "list",
    "lines",
    "cssbeautify",
    "jsbeautify",
    "collapsebraces",
    "attrs",
)


def _load_strategies():
    """STRATEGIES is created at the end of this file.

    Returns:
        mapping: A mapping of strategy names to strategy class.
    """
    strategies = {}
    for entry_point in iter_entry_points("grizzly_reduce_strategies"):
        try:
            strategy_cls = entry_point.load()
            strategy_cls.sanity_check_cls_attrs()
            assert (
                strategy_cls.name == entry_point.name
            ), "entry_point name mismatch, check setup.py and %s.name" % (
                strategy_cls.__name__,
            )
        except Exception as exc:  # pylint: disable=broad-except
            LOG.debug("error loading strategy type %s: %s", entry_point.name, exc)
            continue
        strategies[entry_point.name] = strategy_cls
    for strategy in DEFAULT_STRATEGIES:
        assert (
            strategy in strategies
        ), "Unknown entry in DEFAULT_STRATEGIES: %s (STRATEGIES: [%s])" % (
            strategy,
            ",".join(strategies),
        )
    return MappingProxyType(strategies)


def _contains_dd(path):
    data = path.read_bytes()
    return b"DDBEGIN" in data and b"DDEND" in data


class Strategy(ABC):
    """A strategy is a procedure for repeatedly running a testcase to find the smallest
    equivalent test.

    Implementors must define these class attributes:

    Class Attributes:
        name (str): The strategy name.
    """

    name = None

    def __init__(self, testcases):
        """Initialize strategy instance.

        Arguments:
            testcases (list(grizzly.common.storage.TestCase)):
                List of testcases to reduce. The object does not take ownership of the
                testcases.
        """
        self._tried = set()  # set of tuple(tuple(str(Path), SHA512))
        self._testcase_root = Path(mkdtemp(prefix="tc_", dir=grz_tmp("reduce")))
        self.dump_testcases(testcases)

    def _calculate_testcase_hash(self):
        """Calculate hashes of all files in testcase root.

        Returns:
            tuple(tuple(str, str)): A tuple of 2-tuples mapping str(Path) to SHA-512 of
                                    each file in testcase root.
        """
        result = []
        for path in self._testcase_root.glob("**/*"):
            if path.is_file():
                tf_hash = sha512()
                tf_hash.update(path.read_bytes())
                result.append(
                    (str(path.relative_to(self._testcase_root)), tf_hash.digest())
                )
        result = tuple(sorted(result))

        if LOG.getEffectiveLevel() == DEBUG:
            print_hash = sha512()
            print_hash.update(repr(result).encode("utf-8", errors="surrogateescape"))
            in_tried = result in self._tried
            LOG.debug(
                "Testcase hash: %s (%sin cache)",
                print_hash.hexdigest()[:32],
                "" if in_tried else "not ",
            )

        return result

    def update_tried(self, tried):
        """Update the list of tried testcase/hash sets. Testcases are hashed with
        SHA-512 and digested to bytes (`hashlib.sha512(testcase).digest()`)

        Arguments:
            tried (iterable(tuple(tuple(str, str)))): Set of already tried testcase
                                                      hashes.

        Returns:
            None
        """
        self._tried.update(frozenset(tried))

    def get_tried(self):
        """Return the set of tried testcase hashes. Testcases are hashed with SHA-512
        and digested to bytes (`hashlib.sha512(testcase).digest()`)

        Returns:
            frozenset(tuple(tuple(str, str))): Testcase hashes.
        """
        return frozenset(self._tried)

    def dump_testcases(self, testcases, recreate_tcroot=False):
        """Dump a testcase list to the testcase root on disk.

        Arguments:
            testcases (list(grizzly.common.storage.TestCase)): list of testcases to dump
            recreate_tcroot (bool): if True, delete testcase root and recreate it before
                                    dumping

        Returns:
            None
        """
        if recreate_tcroot:
            rmtree(str(self._testcase_root))
            self._testcase_root.mkdir()
        for idx, testcase in enumerate(testcases):
            LOG.debug("Extracting testcase %d/%d", idx + 1, len(testcases))
            testpath = self._testcase_root / ("%03d" % (idx,))
            testcase.dump(str(testpath), include_details=True)

    @classmethod
    def sanity_check_cls_attrs(cls):
        """Sanity check the strategy class implementation.

        This should assert that any required class attributes are defined and correct.

        Raises:
            AssertionError: Any required class attributes are missing or wrong type.

        Returns:
            None
        """
        assert isinstance(cls.name, str)

    @abstractmethod
    def __iter__(self):
        """Iterate over potential reductions of testcases according to this strategy.

        The caller should evaluate each reduction yielded, and call `update` with the
        result. The caller owns the testcases yielded, and should call `cleanup` for
        each.

        Yields:
            list(grizzly.common.storage.TestCase): list of testcases with reduction
                                                   applied
        """

    @abstractmethod
    def update(self, success, served=None):
        """Inform the strategy whether or not the last reduction yielded was good.

        Arguments:
            success (bool): Whether or not the last reduction was acceptable.
            served (list(list(str))): The list of served files for each testcase in the
                                      last reduction.

        Returns:
            None
        """

    def __enter__(self):
        """Enter a runtime context that will automatically call `cleanup` on exit.

        Returns:
            Strategy: self
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the runtime context. `cleanup` is called.

        Arguments:
            exc_type (type or None): Type of exception object currently raised.
            exc_val (Exception or None): Exception object currently raised.
            exc_tb (traceback or None): Traceback for currently raised exception.

        Returns:
            None
        """
        self.cleanup()

    def cleanup(self):
        """Destroy all resources held by the strategy.

        Returns:
            None
        """
        rmtree(str(self._testcase_root))

    def purge_unserved(self, testcases, served):
        """Given the testcase list yielded and list of what was served, purge
        everything in testcase root to hold only what was served.

        Arguments:
            testcases (list(grizzly.common.storage.TestCase): testcases last replayed
            served (list(list(str))): list of files served for each testcase.

        Returns:
            bool: True if anything was purged
        """
        LOG.debug("purging from %d testcases", len(testcases))
        anything_purged = False
        while len(served) < len(testcases):
            LOG.debug(
                "not all %d testcases served (%d served), popping one",
                len(testcases),
                len(served),
            )
            testcases.pop().cleanup()
            anything_purged = True
        remove_testcases = []
        for idx, (testcase, tc_served) in enumerate(zip(testcases, served)):
            LOG.debug("testcase %d served %r", idx, tc_served)
            if testcase.landing_page not in tc_served:
                LOG.debug("landing page %r not served", testcase.landing_page)
                remove_testcases.append(idx)
                anything_purged = True
            else:
                size_before = testcase.data_size
                testcase.purge_optional(tc_served)
                anything_purged = anything_purged or testcase.data_size != size_before
        for idx in reversed(remove_testcases):
            testcases.pop(idx).cleanup()
        self.dump_testcases(testcases, recreate_tcroot=True)
        return anything_purged


STRATEGIES = _load_strategies()



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Unit tests for `grizzly.reduce.crash` and `grizzly.reduce.bucket`."""
from copy import deepcopy
from logging import getLogger

import pytest

from .bucket import main as bucket_main
from .crash import main as crash_main

LOG = getLogger(__name__)
pytestmark = pytest.mark.usefixtures("tmp_path_fm_config")


@pytest.mark.parametrize(
    "arg_sig, arg_tool, crash_bucket, result_sig, result_tool",
    [
        # crash id is downloaded and core.main is called with the path
        (None, None, None, None, "test-tool"),
        # crash in bucket, bucket will be downloaded and passed to --sig
        (None, None, 789, "test_sig.json", "test-tool"),
        # crash in bucket, --sig is respected
        ("arg_sig.json", None, 789, "arg_sig.json", "test-tool"),
        # --tool respected
        (None, "tool2", None, None, "tool2"),
    ],
)
def test_crash_main(mocker, arg_sig, arg_tool, crash_bucket, result_sig, result_tool):
    """tests for `grizzly.reduce.crash.main`"""
    mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    crash = mocker.patch("grizzly.reduce.crash.CrashEntry", autospec=True)
    bucket = mocker.patch("grizzly.reduce.crash.Bucket", autospec=True)
    mocker.patch("grizzly.reduce.crash.ReduceManager", autospec=True)
    crash.return_value.testcase_path.return_value = "test_path.zip"
    crash.return_value.bucket = crash_bucket
    crash.return_value.testcase_quality = 5
    crash.return_value.tool = "test-tool"
    bucket.return_value.signature_path.return_value = "test_sig.json"

    args = mocker.Mock(
        input=12345,
        sig=arg_sig,
        tool=arg_tool,
    )
    crash_main(args)
    assert args.input == "test_path.zip"
    assert args.sig == result_sig
    assert args.tool == result_tool


@pytest.mark.parametrize(
    "mgr_exit_code, pre_quality, post_quality",
    [
        (0, 5, 1),
        (1, 5, 9),
        (2, 6, 5),
        (3, 4, 5),
        (3, 0, 0),
        (4, 6, 5),
        (5, 5, 10),
    ],
)
def test_crash_main_quality(mocker, mgr_exit_code, pre_quality, post_quality):
    """test that quality is updated"""
    mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    crash = mocker.patch("grizzly.reduce.crash.CrashEntry", autospec=True)
    mocker.patch("grizzly.reduce.crash.Bucket", autospec=True)
    mgr = mocker.patch("grizzly.reduce.crash.ReduceManager", autospec=True)
    crash.return_value.testcase_path.return_value = "test_path.zip"
    crash.return_value.testcase_quality = pre_quality
    crash.return_value.bucket = None
    crash.return_value.tool = "test-tool"
    mgr.main.return_value = mgr_exit_code

    args = mocker.Mock(
        input=12345,
        sig=None,
        tool=None,
    )
    assert crash_main(args) == mgr_exit_code
    assert crash.return_value.testcase_quality == post_quality


@pytest.mark.parametrize(
    "crashes, main_exit_codes, result, arg_sig, arg_tool",
    [
        # no crashes -> success
        ([], [], 0, None, None),
        # 1 crash fails -> no success
        ([(123, "test-tool")], [1], 1, None, None),
        # second of 3 succeeds -> success
        ([(123, "test-tool"), (456, "test-tool2")], [1, 0], 0, None, None),
        # --sig is respected
        ([(123, "test-tool")], [0], 0, "test_sig2.json", None),
        # --tool is respected
        ([(123, "test-tool")], [0], 0, None, "test-tool-arg"),
        # abort in crash main should also abort bucket main
        ([(123, "test-tool")], [3], 3, None, None),
    ],
)
def test_bucket_main(mocker, crashes, main_exit_codes, result, arg_sig, arg_tool):
    """tests for `grizzly.reduce.bucket.main`"""
    call_args = []

    def copy_args(args):
        call_args.append(deepcopy(args))
        return main_exit_codes[main.call_count - 1]

    mocker.patch("grizzly.common.fuzzmanager.Collector", autospec=True)
    bucket = mocker.patch("grizzly.reduce.bucket.Bucket", autospec=True)
    main = mocker.patch(
        "grizzly.reduce.bucket.crash_main", autospec=True, side_effect=copy_args
    )
    bucket.return_value.signature_path.return_value = "test_sig.json"
    bucket.return_value.iter_crashes.return_value = [
        mocker.Mock(crash_id=crash, tool=tool) for crash, tool in crashes
    ]

    args = mocker.Mock(
        input=789,
        sig=arg_sig,
        tool=arg_tool,
    )
    assert bucket_main(args) == result
    assert main.call_count == len(main_exit_codes)
    for idx, (crash, _tool) in enumerate(crashes[: main.call_count]):
        assert call_args[idx].input == crash
        if arg_tool is not None:
            assert call_args[idx].tool == arg_tool
        else:
            assert call_args[idx].tool is None
        if arg_sig is not None:
            assert call_args[idx].sig == arg_sig
        else:
            assert call_args[idx].sig == "test_sig.json"



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from pytest import mark, raises

from .args import CommonArgs, GrizzlyArgs


def test_common_args_01(capsys, mocker):
    """test CommonArgs.parse_args()"""
    mocker.patch("grizzly.args.scan_plugins", autospec=True, return_value=[])
    # test help
    with raises(SystemExit):
        CommonArgs().parse_args(argv=["-h"])
    assert "For addition help check out the wiki" in capsys.readouterr()[0]
    # test empty args
    with raises(SystemExit):
        CommonArgs().parse_args(argv=[])
    assert "the following arguments are required: binary" in capsys.readouterr()[-1]


def test_common_args_01a(capsys, mocker, tmp_path):
    """test CommonArgs.parse_args()"""
    mocker.patch("grizzly.args.scan_plugins", autospec=True, return_value=["targ1"])
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    # test with missing bin
    with raises(SystemExit):
        CommonArgs().parse_args(argv=["missing-bin"])
    assert "error: file not found: 'missing-bin'" in capsys.readouterr()[-1]
    # test success
    CommonArgs().parse_args(argv=[str(fake_bin), "--platform", "targ1"])
    # test invalid extension
    with raises(SystemExit):
        CommonArgs().parse_args(argv=[str(fake_bin), "--extension", str(fake_bin)])
    assert "error: Extension must be a folder or .xpi" in capsys.readouterr()[-1]


@mark.parametrize(
    "args, msg, targets",
    [
        # test no installed platforms
        ([], "error: No Platforms (Targets) are installed", []),
        # test invalid ignore value
        (["--ignore", "bad"], "error: Unrecognized ignore value 'bad'", ["targ1"]),
        # test invalid log level
        (["--log-level", "bad"], "error: Invalid log-level 'bad'", ["targ1"]),
        # test invalid log limit
        (["--log-limit", "-1"], "error: --log-limit must be >= 0", ["targ1"]),
        # test invalid memory limit
        (["--memory", "-1"], "error: --memory must be >= 0", ["targ1"]),
        # test invalid relaunch value
        (["--relaunch", "0"], "error: --relaunch must be >= 1", ["targ1"]),
        # test missing extension
        (["--extension", "missing"], "error: 'missing' does not exist", ["targ1"]),
        # test invalid platform/target
        (["--platform", "bad"], "error: Platform 'bad' not installed", ["targ1"]),
        # test invalid prefs file
        (
            ["--platform", "targ1", "--prefs", "bad"],
            "error: --prefs file not found",
            ["targ1"],
        ),
        # test invalid time-limit
        (
            ["--platform", "targ1", "--time-limit", "-1"],
            "error: --time-limit must be >= 1",
            ["targ1"],
        ),
        # test invalid timeout
        (
            ["--platform", "targ1", "--timeout", "-1"],
            "error: --timeout must be >= 1",
            ["targ1"],
        ),
        # test invalid tool usage
        (
            ["--platform", "targ1", "--tool", "x"],
            "error: --tool can only be given with --fuzzmanager",
            ["targ1"],
        ),
        # test enabling both rr and Valgrind
        (
            ["--platform", "targ1", "--rr", "--valgrind"],
            "error: --rr and --valgrind are mutually exclusive",
            ["targ1"],
        ),
        # test rr on unsupported platform
        (
            ["--platform", "targ1", "--rr"],
            "error: --rr is only supported on Linux",
            ["targ1"],
        ),
        # test Valgrind on unsupported platform
        (
            ["--platform", "targ1", "--valgrind"],
            "error: --valgrind is only supported on Linux",
            ["targ1"],
        ),
        # test Xvfb on unsupported platform
        (
            ["--platform", "targ1", "--xvfb"],
            "error: --xvfb is only supported on Linux",
            ["targ1"],
        ),
    ],
)
def test_common_args_02(capsys, mocker, tmp_path, args, msg, targets):
    """test CommonArgs.parse_args()"""
    mocker.patch("grizzly.args.scan_plugins", autospec=True, return_value=targets)
    mocker.patch("grizzly.args.system", autospec=True, return_value="foo")
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    with raises(SystemExit):
        CommonArgs().parse_args(argv=[str(fake_bin)] + args)
    assert msg in capsys.readouterr()[-1]


def test_grizzly_args_01(mocker, tmp_path):
    """test GrizzlyArgs.parse_args() - success"""
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    mocker.patch(
        "grizzly.args.scan_plugins",
        autospec=True,
        side_effect=(["targ1"], ["adpt1"], ["targ1"], ["adpt1"]),
    )
    assert GrizzlyArgs().parse_args(
        argv=[str(fake_bin), "adpt1", "--platform", "targ1"]
    )


def test_grizzly_args_02(capsys, mocker):
    """test GrizzlyArgs.parse_args() - handling binary"""
    mocker.patch("grizzly.args.scan_plugins", autospec=True, return_value=["blah"])
    # test missing required args
    with raises(SystemExit):
        GrizzlyArgs().parse_args(argv=[])
    err = capsys.readouterr()[-1]
    assert "the following arguments are required: binary, adapter" in err
    # test missing binary
    with raises(SystemExit):
        GrizzlyArgs().parse_args(argv=["missing_bin", "adapter"])
    assert "error: file not found: 'missing_bin'" in capsys.readouterr()[-1]


def test_grizzly_args_03(capsys, mocker, tmp_path):
    """test GrizzlyArgs.parse_args() - handling Adapter"""
    scan_plugins = mocker.patch("grizzly.args.scan_plugins", autospec=True)
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    # no adapters installed
    scan_plugins.side_effect = (["targ1"], [], ["targ1"], [])
    with raises(SystemExit):
        GrizzlyArgs().parse_args(argv=[str(fake_bin), "adpt", "--platform", "targ1"])
    assert "error: No Adapters are installed" in capsys.readouterr()[-1]
    # invalid adapter name
    scan_plugins.side_effect = (["targ1"], ["a1", "a2"], ["targ1"], ["a1", "a2"])
    with raises(SystemExit):
        GrizzlyArgs().parse_args(argv=[str(fake_bin), "missing", "--platform", "targ1"])
    assert "error: Adapter 'missing' is not installed" in capsys.readouterr()[-1]


@mark.parametrize(
    "args, msg",
    [
        # test invalid collect value
        (["--collect", "0"], "error: --collect must be greater than 0"),
        # test enabling both fuzzmanager and s3-fuzzmanager reporters
        (
            ["--fuzzmanager", "--s3-fuzzmanager"],
            "error: --fuzzmanager and --s3-fuzzmanager are mutually exclusive",
        ),
        # test missing input
        (["--input", "missing"], "error: 'missing' does not exist"),
        # test invalid limit value
        (["--limit", "-1"], "error: --limit must be >= 0 (0 = no limit)"),
        # test tool
        (
            ["--tool", "x"],
            "error: --tool can only be given with --fuzzmanager/--s3-fuzzmanager",
        ),
    ],
)
def test_grizzly_args_04(capsys, mocker, tmp_path, args, msg):
    """test CommonArgs.parse_args()"""
    mocker.patch(
        "grizzly.args.scan_plugins",
        autospec=True,
        side_effect=["targ1", "adpt", "targ1", "adpt"],
    )
    fake_bin = tmp_path / "fake.bin"
    fake_bin.touch()
    with raises(SystemExit):
        GrizzlyArgs().parse_args(
            argv=[str(fake_bin), "adpt", "--platform", "targ1"] + args
        )
    assert msg in capsys.readouterr()[-1]



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""
unit tests for grizzly.Session
"""
from pytest import mark, raises

from sapphire import SERVED_ALL, SERVED_NONE, SERVED_TIMEOUT, Sapphire

from .adapter import Adapter
from .common import Report, Reporter, RunResult, Status
from .session import LogOutputLimiter, Session, SessionError
from .target import Target, TargetLaunchError


class SimpleAdapter(Adapter):
    def __init__(self, use_harness, remaining=None):
        super().__init__("simple")
        self.remaining = remaining
        self._use_harness = use_harness

    def setup(self, input_path, _server_map):
        if self._use_harness:
            self.enable_harness()
        self.fuzz["input"] = input_path

    def generate(self, testcase, _server_map):
        assert testcase.adapter_name == self.name
        testcase.input_fname = self.fuzz["input"]
        testcase.add_from_data("test", testcase.landing_page)
        if self.remaining is not None:
            assert self.remaining > 0
            self.remaining -= 1


@mark.parametrize(
    "harness, profiling, coverage, relaunch, iters",
    [
        # with harness, single iteration
        (True, False, False, 1, 1),
        # with harness, 10 iterations relaunch every iteration
        (True, False, False, 1, 10),
        # with harness, 10 iterations relaunch every other iteration
        (True, False, False, 2, 10),
        # with harness, 10 iterations no relaunches
        (True, False, False, 10, 10),
        # no harness, single iteration
        (False, False, False, 1, 1),
        # no harness, 10 iterations
        (False, False, False, 1, 10),
        # test enable profiling
        (True, True, False, 10, 10),
        # test Session.dump_coverage()
        (True, True, True, 2, 2),
    ],
)
def test_session_01(mocker, tmp_path, harness, profiling, coverage, relaunch, iters):
    """test Session with typical fuzzer Adapter"""
    Status.PATH = str(tmp_path)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    prefs = tmp_path / "prefs.js"
    prefs.touch()
    target = mocker.Mock(spec=Target, launch_timeout=30, prefs=str(prefs))
    target.log_size.return_value = 1000
    target.monitor.launches = 1
    # avoid shutdown delay
    target.monitor.is_healthy.return_value = False
    # calculate if the target is 'closed' based on relaunch
    type(target).closed = mocker.PropertyMock(
        side_effect=((x % relaunch == 0) for x in range(iters))
    )
    with Session(
        SimpleAdapter(harness),
        None,
        server,
        target,
        coverage=coverage,
        enable_profiling=profiling,
        relaunch=relaunch,
    ) as session:
        server.serve_path = lambda *a, **kv: (
            SERVED_ALL,
            [session.iomanager.page_name(offset=-1)],
        )
        session.run([], 10, input_path="file.bin", iteration_limit=iters)
        assert session.status.iteration == iters
        assert session.status.test_name == "file.bin"
        assert target.close.call_count == iters / relaunch
        assert target.detect_failure.call_count == iters
        assert target.handle_hang.call_count == 0
        if profiling:
            assert any(session.status.profile_entries())
        else:
            assert not any(session.status.profile_entries())
        if coverage:
            assert target.dump_coverage.call_count == iters
        else:
            assert target.dump_coverage.call_count == 0


@mark.parametrize(
    "harness, relaunch, remaining",
    [
        # no harness, 1 iteration
        (False, 1, 1),
        # no harness, 5 iterations, relaunch every iteration
        (False, 1, 5),
        # harness, 1 iteration
        (True, 1, 1),
        # harness, 10 iterations, relaunch every other iteration
        (True, 2, 10),
        # harness, 10 iterations
        (True, 10, 10),
    ],
)
def test_session_02(tmp_path, mocker, harness, relaunch, remaining):
    """test Session with playback Adapter"""
    Status.PATH = str(tmp_path)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target, launch_timeout=30, prefs=None)
    # calculate if the target is 'closed' based on relaunch
    type(target).closed = mocker.PropertyMock(
        side_effect=((x % relaunch == 0) for x in range(remaining))
    )
    # avoid shutdown delay
    target.monitor.is_healthy.return_value = False
    # set target.log_size to test warning code path
    target.log_size.return_value = Session.TARGET_LOG_SIZE_WARN + 1
    with Session(
        SimpleAdapter(harness, remaining=remaining),
        None,
        server,
        target,
        relaunch=relaunch,
    ) as session:
        server.serve_path = lambda *a, **kv: (
            SERVED_ALL,
            [session.iomanager.page_name(offset=-1)],
        )
        session.run([], 10)
        assert session.status.iteration == remaining
        assert session.status.test_name is None
        assert target.detect_failure.call_count == remaining
        assert target.handle_hang.call_count == 0


@mark.parametrize(
    "harness, report_size, relaunch, iters",
    [
        # with harness, collect 1 test case
        (True, 1, 1, 1),
        # with harness, collect 2 test cases
        (True, 2, 2, 2),
        # with harness, collect 2 test cases,
        (True, 2, 3, 3),
        # with harness, collect 3 test cases, relaunch 1
        (True, 3, 1, 3),
        # without harness, collect 1 test case
        (False, 1, 1, 1),
        # without harness, collect 1 test case, 3 iterations
        (False, 1, 1, 3),
    ],
)
def test_session_03(mocker, tmp_path, harness, report_size, relaunch, iters):
    """test Session - detecting failure"""
    Status.PATH = str(tmp_path)
    adapter = SimpleAdapter(harness)
    reporter = mocker.Mock(spec=Reporter)
    report = mocker.Mock(spec=Report, major="major123", minor="minor456")
    report.crash_info.createShortSignature.return_value = "[@ sig]"
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target, launch_timeout=30, prefs=None)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_NONE = Target.RESULT_NONE
    target.monitor.launches = 1
    # avoid shutdown delay
    target.monitor.is_healthy.return_value = False
    # calculate if the target is 'closed' based on relaunch
    type(target).closed = mocker.PropertyMock(
        side_effect=((x % relaunch == 0) for x in range(iters))
    )
    # failure is on final iteration
    target.detect_failure.side_effect = [
        Target.RESULT_NONE for x in range(iters - 1)
    ] + [Target.RESULT_FAILURE]
    target.log_size.return_value = 1
    target.create_report.return_value = report
    with Session(
        adapter, reporter, server, target, relaunch=relaunch, report_size=report_size
    ) as session:
        server.serve_path = lambda *a, **kv: (
            SERVED_ALL,
            [session.iomanager.page_name(offset=-1)],
        )
        session.run([], 10, input_path="file.bin", iteration_limit=iters)
        assert reporter.submit.call_count == 1
        assert len(reporter.submit.call_args[0][0]) == min(report_size, relaunch)
        assert reporter.submit.call_args[0][1].major == "major123"


def test_session_04(mocker, tmp_path):
    """test Adapter creating invalid test case"""

    class FuzzAdapter(Adapter):
        def generate(self, _testcase, _server_map):
            pass

    Status.PATH = str(tmp_path)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_path.return_value = (SERVED_NONE, [])
    target = mocker.Mock(spec=Target, launch_timeout=30, prefs=None)
    target.monitor.launches = 1
    with Session(FuzzAdapter("fuzz"), None, server, target) as session:
        with raises(SessionError, match="Test case is missing landing page"):
            session.run([], 10)


def test_session_05(mocker, tmp_path):
    """test Target not requesting landing page"""
    Status.PATH = str(tmp_path)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    server.serve_path.return_value = (SERVED_TIMEOUT, [])
    target = mocker.Mock(spec=Target, launch_timeout=30, prefs=None)
    target.monitor.launches = 1
    with Session(SimpleAdapter(False), None, server, target) as session:
        with raises(SessionError, match="Please check Adapter and Target"):
            session.run([], 10)


@mark.parametrize(
    "harness, report_size",
    [
        # with harness, collect 1 test case
        (True, 1),
        # with harness, collect 2 test cases
        (True, 2),
        # without harness, collect 1 test case
        (False, 1),
    ],
)
def test_session_06(mocker, tmp_path, harness, report_size):
    """test Session - handle Target delayed failures"""
    Status.PATH = str(tmp_path)
    reporter = mocker.Mock(spec=Reporter)
    report = mocker.Mock(spec=Report, major="major123", minor="minor456")
    report.crash_info.createShortSignature.return_value = "[@ sig]"
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target, launch_timeout=30, prefs=None)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_NONE = Target.RESULT_NONE
    target.monitor.launches = 1
    type(target).closed = mocker.PropertyMock(side_effect=(True, False))
    target.detect_failure.side_effect = (Target.RESULT_NONE, Target.RESULT_FAILURE)
    target.log_size.return_value = 1
    target.create_report.return_value = report
    with Session(
        SimpleAdapter(harness),
        reporter,
        server,
        target,
        relaunch=2,
        report_size=report_size,
    ) as session:
        server.serve_path.side_effect = (
            (SERVED_ALL, [session.iomanager.page_name()]),
            (SERVED_NONE, []),
        )
        session.run([], 10, iteration_limit=2)
        assert reporter.submit.call_count == 1
        assert len(reporter.submit.call_args[0][0]) == 1
        assert reporter.submit.call_args[0][1].major == "major123"


@mark.parametrize(
    "srv_results, target_result, ignored, results",
    [
        # delayed startup crash
        (SERVED_NONE, Target.RESULT_FAILURE, 0, 1),
        # startup hang/unresponsive
        (SERVED_TIMEOUT, Target.RESULT_NONE, 1, 0),
    ],
)
def test_session_07(mocker, tmp_path, srv_results, target_result, ignored, results):
    """test Session.run() - initial test case was not served"""
    Status.PATH = str(tmp_path)
    report = mocker.Mock(spec=Report, major="major123", minor="minor456")
    report.crash_info.createShortSignature.return_value = "[@ sig]"
    reporter = mocker.Mock(spec=Reporter)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target, closed=True, launch_timeout=30, prefs=None)
    target.RESULT_FAILURE = Target.RESULT_FAILURE
    target.RESULT_NONE = Target.RESULT_NONE
    target.monitor.launches = 1
    target.detect_failure.side_effect = (target_result,)
    target.create_report.return_value = report
    with Session(SimpleAdapter(False), reporter, server, target) as session:
        server.serve_path.return_value = (srv_results, [])
        with raises(SessionError, match="Please check Adapter and Target"):
            session.run([], 10, iteration_limit=2)
        assert session.status.iteration == 1
        assert session.status.results == results
        assert session.status.ignored == ignored
        assert reporter.submit.call_count == results
        assert target.detect_failure.call_count == results
        assert target.handle_hang.call_count == ignored


def test_session_08(tmp_path, mocker):
    """test Session.run() ignoring failures"""
    Status.PATH = str(tmp_path)
    result = RunResult([], 0.1, status=RunResult.IGNORED)
    result.attempted = True
    runner = mocker.patch("grizzly.session.Runner", autospec=True)
    runner.return_value.run.return_value = result
    mocker.patch("grizzly.session.TestFile", autospec=True)
    adapter = mocker.Mock(spec=Adapter, remaining=None)
    adapter.IGNORE_UNSERVED = False
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target, prefs=None)
    target.monitor.launches = 1
    with Session(adapter, None, server, target) as session:
        session.run([], 10, iteration_limit=1)
        assert runner.return_value.run.call_count == 1
        assert adapter.on_served.call_count == 1
        assert adapter.on_timeout.call_count == 0
        assert target.create_report.call_count == 0
        assert session.status.iteration == 1
        assert session.status.results == 0
        assert session.status.ignored == 1


def test_session_09(tmp_path, mocker):
    """test Session.run() handle TargetLaunchError"""
    Status.PATH = str(tmp_path)
    report = mocker.Mock(spec=Report, major="major123", minor="minor456")
    runner = mocker.patch("grizzly.session.Runner", autospec=True)
    runner.return_value.launch.side_effect = TargetLaunchError("test", report)
    mocker.patch("grizzly.session.TestFile", autospec=True)
    adapter = mocker.Mock(spec=Adapter, remaining=None)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target)
    target.monitor.launches = 1
    with Session(adapter, mocker.Mock(spec=Reporter), server, target) as session:
        with raises(TargetLaunchError, match="test"):
            session.run([], 10, iteration_limit=2)
        assert session.status.iteration == 1
        assert session.status.results == 1
        assert session.status.ignored == 0


def test_session_10(tmp_path, mocker):
    """test Session.run() report hang"""
    Status.PATH = str(tmp_path)
    result = RunResult([], 60.0, status=RunResult.FAILED, timeout=True)
    result.attempted = True
    runner = mocker.patch("grizzly.session.Runner", autospec=True)
    runner.return_value.run.return_value = result
    mocker.patch("grizzly.session.TestFile", autospec=True)
    adapter = mocker.Mock(spec=Adapter, remaining=None)
    report = mocker.Mock(spec=Report, major="major123", minor="minor456")
    reporter = mocker.Mock(spec=Reporter)
    server = mocker.Mock(spec=Sapphire, port=0x1337)
    target = mocker.Mock(spec=Target, prefs=None)
    target.monitor.launches = 1
    target.create_report.return_value = report
    with Session(adapter, reporter, server, target) as session:
        session.run([], 10, iteration_limit=1)
        assert runner.return_value.run.call_count == 1
        assert adapter.on_served.call_count == 0
        assert adapter.on_timeout.call_count == 1
        assert target.create_report.call_count == 1
        assert reporter.submit.call_count == 1
        assert session.status.iteration == 1
        assert session.status.results == 1
        assert session.status.ignored == 0


def test_log_output_limiter_01(mocker):
    """test LogOutputLimiter.ready() not ready"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=10, delta_multiplier=2)
    assert lol._delay == 10
    assert lol._iterations == 1
    assert lol._launches == 1
    assert lol._multiplier == 2
    assert lol._time == 1.0
    assert not lol._verbose
    fake_time.return_value = 1.1
    assert not lol.ready(0, 0)
    assert lol._iterations == 1
    assert lol._launches == 1
    assert lol._time == 1.0
    lol._verbose = True
    assert lol.ready(0, 0)


def test_log_output_limiter_02(mocker):
    """test LogOutputLimiter.ready() due to iterations"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=10, delta_multiplier=2)
    fake_time.return_value = 1.1
    lol._launches = 2
    assert lol.ready(1, 1)
    assert lol._iterations == 2
    assert lol._launches == 2
    assert lol._time == 1.1


def test_log_output_limiter_03(mocker):
    """test LogOutputLimiter.ready() due to launches"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=10, delta_multiplier=2)
    lol._iterations = 4
    assert lol.ready(3, 1)
    assert lol._launches == 2
    assert lol._iterations == 4
    assert lol._time == 1.0


def test_log_output_limiter_04(mocker):
    """test LogOutputLimiter.ready() due to time"""
    fake_time = mocker.patch("grizzly.session.time", autospec=True)
    fake_time.return_value = 1.0
    lol = LogOutputLimiter(delay=1, delta_multiplier=2)
    lol._iterations = 4
    fake_time.return_value = 2.0
    assert lol.ready(3, 0)
    assert lol._iterations == 4
    assert lol._launches == 1
    assert lol._time == 2.0



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from os.path import exists, isfile

from ..args import CommonArgs


class ReplayArgs(CommonArgs):
    def __init__(self):
        super().__init__()
        self.parser.add_argument(
            "input",
            help="Accepted input includes: "
            "1) A directory containing testcase data. "
            "2) A directory with one or more subdirectories containing testcase data. "
            "3) A zip archive containing testcase data or subdirectories containing"
            " testcase data. "
            "4) A single file to be used as a testcase. "
            "When using a directory it must contain a 'test_info.json' file.",
        )

        replay_args = self.parser.add_argument_group("Replay Arguments")
        replay_args.add_argument(
            "--any-crash",
            action="store_true",
            help="Any crash is interesting, not only crashes which match the original"
            " signature.",
        )
        replay_args.add_argument(
            "--idle-delay",
            type=int,
            default=30,
            help="Number of seconds to wait before polling for idle"
            " (default: %(default)s)",
        )
        replay_args.add_argument(
            "--idle-threshold",
            type=int,
            default=0,
            help="CPU usage threshold to mark the process as idle (default: disabled)",
        )
        replay_args.add_argument(
            "-l",
            "--logs",
            help="Location to save logs. If the path exists it must be empty, if it "
            "does not exist it will be created.",
        )
        replay_args.add_argument(
            "--min-crashes",
            type=int,
            default=1,
            help="Require the testcase to crash n times before accepting the result."
            " Helpful for intermittent testcases (default: %(default)sx)",
        )
        replay_args.add_argument(
            "--no-harness",
            action="store_true",
            help="Don't use the harness for redirection. Implies '--relaunch=1'.",
        )
        replay_args.add_argument(
            "--repeat",
            type=int,
            default=1,
            help="Run the testcase n times."
            " Helpful for intermittent testcases (default: %(default)sx)",
        )
        replay_args.add_argument("--sig", help="Signature (JSON) file to match.")
        replay_args.add_argument(
            "--test-index",
            type=int,
            nargs="+",
            help="Select a testcase to run when multiple testcases are loaded. "
            "Testscases are ordered oldest to newest. Indexing is 0 based. "
            "0 == Oldest, n-1 == Newest (default: run all testcases)",
        )

        self.reporter_grp.add_argument(
            "--include-test",
            action="store_true",
            help="Include the testcase when reporting results.",
        )

    def sanity_check(self, args):
        super().sanity_check(args)

        if args.any_crash and args.sig is not None:
            self.parser.error("signature is ignored when running with --any-crash")

        if args.idle_threshold and args.idle_delay <= 0:
            self.parser.error("--idle-delay value must be positive")

        if "input" not in self._sanity_skip and not exists(args.input):
            self.parser.error("%r does not exist" % (args.input,))

        if args.min_crashes < 1:
            self.parser.error("--min-crashes value must be positive")

        if args.repeat < 1:
            self.parser.error("--repeat value must be positive")

        if args.sig is not None and not isfile(args.sig):
            self.parser.error("signature file not found: %r" % (args.sig,))



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly status"""
# pylint: disable=protected-access

from multiprocessing import Event, Process
from os import remove, stat
from os.path import isfile
from time import sleep, time

from .status import Status


def test_status_01(tmp_path):
    """test Status.start()"""
    Status.PATH = str(tmp_path)
    status = Status.start()
    assert status is not None
    assert status.data_file is not None
    assert isfile(status.data_file)
    assert stat(status.data_file).st_size > 0
    assert status.start_time > 0
    assert status.timestamp >= status.start_time
    assert int(status.duration) == 0
    assert status.ignored == 0
    assert status.iteration == 0
    assert status.log_size == 0
    assert status.rate == 0
    assert status.results == 0
    assert status.pid is not None
    assert not status._enable_profiling
    assert not status._profiles


def test_status_02(tmp_path):
    """test Status.cleanup()"""
    Status.PATH = str(tmp_path)
    status = Status.start()
    dfile = status.data_file
    status.cleanup()
    assert status.data_file is None
    assert not isfile(dfile)
    # call 2nd time
    status.cleanup()
    # missing data file
    status = Status.start()
    remove(status.data_file)
    status.cleanup()


def test_status_03(tmp_path):
    """test Status.report()"""
    Status.PATH = str(tmp_path)
    status = Status.start()
    status.count_result("sig1")
    # try to report before REPORT_FREQ elapses
    assert not status.report()
    # REPORT_FREQ elapses
    status.timestamp = 0
    assert status.report()
    assert status.timestamp > 0
    # force report
    future = int(time()) + 1000
    status.timestamp = future
    assert status.report(force=True)
    assert status.timestamp < future
    status.cleanup()


def test_status_04(tmp_path):
    """test Status.load() failure paths"""
    Status.PATH = str(tmp_path)
    # load no db
    assert Status.load(str(tmp_path / "missing.json")) is None
    # load empty
    bad = tmp_path / "bad.json"
    bad.touch()
    assert Status.load(str(bad)) is None
    # load invalid/incomplete json
    bad.write_bytes(b"{}")
    assert Status.load(str(bad)) is None


def test_status_05(tmp_path):
    """test Status.load()"""
    Status.PATH = str(tmp_path)
    # create simple entry
    status = Status.start(enable_profiling=True)
    status.count_result("sig1")
    status.record("test", 123.45)
    status.report(force=True)
    assert status.results == 1
    loaded = Status.load(status.data_file)
    assert loaded.data_file is None
    assert status.start_time == loaded.start_time
    assert status.timestamp == loaded.timestamp
    assert status.duration == loaded.duration
    assert status.ignored == loaded.ignored
    assert status.iteration == loaded.iteration
    assert status.log_size == loaded.log_size
    assert status.pid == loaded.pid
    assert status.results == loaded.results
    assert not loaded._enable_profiling
    assert "test" in loaded._profiles
    loaded.cleanup()
    assert isfile(status.data_file)
    data_file = status.data_file
    status.cleanup()
    assert not isfile(data_file)


def test_status_06(tmp_path):
    """test Status.loadall()"""
    working_path = tmp_path / "status"
    Status.PATH = str(working_path)
    # missing path
    assert not any(Status.loadall())
    # no status data
    working_path.mkdir()
    assert not any(Status.loadall())
    # add more entries
    for _ in range(5):
        Status.start()
    (working_path / "empty.json").touch()
    assert len(tuple(Status.loadall())) == 5


def test_status_07(tmp_path):
    """test Status.duration and Status.rate calculations"""
    Status.PATH = str(tmp_path)
    status = Status.start()
    status.start_time = 1
    status.timestamp = 2
    status.iteration = 0
    assert status.duration == 1
    assert status.rate == 0
    status.iteration = 1
    assert status.rate == 1
    status.timestamp += 1
    assert status.rate == 0.5


def _client_writer(done, reported, working_path):
    """Used by test_status_08"""
    # NOTE: this must be at the top level to work on Windows
    Status.PATH = working_path
    status = Status.start()
    try:
        while not done.is_set():
            status.iteration += 1
            status.report(force=True)
            # perform two reports before setting flag
            if not reported.is_set() and status.iteration > 1:
                reported.set()
            sleep(0.01)
    finally:
        status.cleanup()


def test_status_08(tmp_path):
    """test Status.loadall() with multiple active reporters"""
    Status.PATH = str(tmp_path)
    done = Event()
    procs = list()
    report_events = list()
    try:
        # launch processes
        for _ in range(5):
            report_events.append(Event())
            procs.append(
                Process(
                    target=_client_writer, args=(done, report_events[-1], Status.PATH)
                )
            )
            procs[-1].start()
        # wait for processes to launch and report
        for has_reported in report_events:
            assert has_reported.wait(timeout=60)
        # collect reports
        reports = tuple(Status.loadall())
        assert len(reports) == len(procs)
        assert max(x.rate for x in reports) > 0
    finally:
        done.set()
        for proc in procs:
            if proc.pid is not None:
                proc.join()
    # verify cleanup
    assert not any(Status.loadall())


def test_status_09(tmp_path):
    """test Status.count_result() and Status.signatures()"""
    Status.PATH = str(tmp_path)
    status = Status.start()
    status.count_result("sig1")
    status.count_result("sig2")
    status.count_result("sig1")
    status.count_result("sig3")
    assert status.results == 4
    found = dict()
    for sig, count in status.signatures():
        found[sig] = count
    assert "sig1" in found
    assert found["sig1"] == 2
    assert "sig2" in found
    assert found["sig2"] == 1
    assert "sig3" in found
    assert found["sig3"] == 1


def test_status_10(tmp_path):
    """test Status.measure() and Status.record() - profiling support"""
    Status.PATH = str(tmp_path)
    # profiling disabled
    status = Status.start(enable_profiling=False)
    status.record("x", 10.1)
    assert not status._profiles
    with status.measure("x"):
        pass
    assert not status._profiles
    status.cleanup()
    # profiling enabled
    status = Status.start(enable_profiling=True)
    assert not status._profiles
    # initial entry
    status.record("test1", 10.1)
    assert "test1" in status._profiles
    assert status._profiles["test1"]["count"] == 1
    assert status._profiles["test1"]["max"] == 10.1
    assert status._profiles["test1"]["min"] == 10.1
    assert status._profiles["test1"]["total"] == 10.1
    entry = next(status.profile_entries())
    assert entry.name == "test1"
    assert entry.count == 1
    assert entry.max == 10.1
    assert entry.min == 10.1
    assert entry.total == 10.1
    # new min
    status.record("test1", 0.4)
    entry = next(status.profile_entries())
    assert entry.count == 2
    assert entry.max == 10.1
    assert entry.min == 0.4
    assert entry.total == 10.5
    # entry
    status.record("test1", 2)
    entry = next(status.profile_entries())
    assert entry.count == 3
    assert entry.max == 10.1
    assert entry.min == 0.4
    assert entry.total == 12.5
    # new max
    status.record("test1", 99.12)
    entry = next(status.profile_entries())
    assert entry.count == 4
    assert entry.max == 99.12
    assert entry.min == 0.4
    assert entry.total == 111.62
    # new name
    status.record("test2", 1)
    assert "test2" in status._profiles
    assert len(status._profiles) == 2
    assert status._profiles["test2"]["count"] == 1
    assert status._profiles["test2"]["max"] == 1
    assert status._profiles["test2"]["min"] == 1
    assert status._profiles["test2"]["total"] == 1
    status.record("test2", 1)
    assert status._profiles["test2"]["count"] == 2
    assert status._profiles["test2"]["max"] == 1
    assert status._profiles["test2"]["min"] == 1
    assert status._profiles["test2"]["total"] == 2
    # test measure
    with status.measure("no-op"):
        pass
    assert len(status._profiles) == 3
    assert "no-op" in status._profiles
    assert len(tuple(status.profile_entries())) == 3



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .iomanager import IOManager, ServerMap
from .reporter import (
    FilesystemReporter,
    FuzzManagerReporter,
    Report,
    Reporter,
    S3FuzzManagerReporter,
)
from .runner import Runner, RunResult
from .status import Status
from .storage import TestCase, TestCaseLoadFailure, TestFile, TestFileExists
from .utils import configure_logging, grz_tmp

__all__ = (
    "FilesystemReporter",
    "FuzzManagerReporter",
    "configure_logging",
    "grz_tmp",
    "IOManager",
    "Report",
    "Reporter",
    "Runner",
    "RunResult",
    "S3FuzzManagerReporter",
    "ServerMap",
    "Status",
    "TestCase",
    "TestCaseLoadFailure",
    "TestFile",
    "TestFileExists",
)
__author__ = "Jesse Schwartzentruber"
__credits__ = ["Jesse Schwartzentruber", "Tyson Smith"]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from itertools import count
from os.path import isfile
from platform import system

from ffpuppet import BrowserTerminatedError, BrowserTimeoutError, FFPuppet
from pytest import mark, raises

from .puppet_target import PuppetTarget
from .target import Target, TargetError, TargetLaunchError, TargetLaunchTimeout


def test_puppet_target_01(mocker, tmp_path):
    """test creating a PuppetTarget"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.return_value.reason = FFPuppet.RC_CLOSED
    fake_ffp.return_value.log_length.return_value = 562
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), None, 300, 25, 5000) as target:
        assert target.closed
        assert not target._remove_prefs
        prefs_file = target.prefs
        assert isfile(prefs_file)
        assert target._remove_prefs
        assert target.detect_failure([]) == Target.RESULT_NONE
        assert target.log_size() == 1124
        fake_ffp.return_value.log_length.assert_any_call("stderr")
        fake_ffp.return_value.log_length.assert_any_call("stdout")
        assert target.monitor is not None
        target.add_abort_token("test")
        assert fake_ffp.return_value.add_abort_token.call_count == 1
        target.save_logs("fake_dest")
        assert fake_ffp.return_value.save_logs.call_count == 1
    assert fake_ffp.return_value.clean_up.call_count == 1
    assert not isfile(prefs_file)
    # with extra args
    with PuppetTarget(str(fake_file), None, 1, 1, 1, rr=True, fake=1) as target:
        pass


def test_puppet_target_02(mocker, tmp_path):
    """test PuppetTarget.launch()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    # test providing prefs.js
    with PuppetTarget(str(fake_file), None, 300, 25, 5000) as target:
        target.prefs = str(fake_file)
        assert not target._remove_prefs
        # launch success
        target.launch("launch_target_page")
        assert fake_ffp.return_value.launch.call_count == 1
        assert fake_ffp.return_value.close.call_count == 0
        target.close()
        # launch timeout
        fake_ffp.reset_mock()
        fake_ffp.return_value.launch.side_effect = BrowserTimeoutError("timeout")
        with raises(TargetLaunchTimeout, match="timeout"):
            target.launch("launch_target_page")
        assert fake_ffp.return_value.save_logs.call_count == 0
        # launch failure
        fake_ffp.reset_mock()
        (tmp_path / "log_stderr.txt").write_text("fake log")
        (tmp_path / "log_stdout.txt").write_text("fake log")
        mocker.patch(
            "grizzly.target.puppet_target.mkdtemp",
            autospec=True,
            return_value=str(tmp_path),
        )
        fake_ffp.return_value.launch.side_effect = BrowserTerminatedError("fail")
        with raises(TargetLaunchError, match="fail"):
            target.launch("launch_target_page")
        assert fake_ffp.return_value.save_logs.call_count == 1


@mark.parametrize(
    "healthy, reason, ignore, result, closes",
    [
        # running as expected - no failures
        (True, None, [], Target.RESULT_NONE, 0),
        # browser process closed
        (False, FFPuppet.RC_CLOSED, [], Target.RESULT_NONE, 1),
        # browser process crashed
        (False, FFPuppet.RC_ALERT, [], Target.RESULT_FAILURE, 1),
        # browser exit with no crash logs
        (False, FFPuppet.RC_EXITED, [], Target.RESULT_NONE, 1),
        # ffpuppet check failed
        (False, FFPuppet.RC_WORKER, [], Target.RESULT_FAILURE, 1),
        # ffpuppet check ignored (memory)
        (False, FFPuppet.RC_WORKER, ["memory"], Target.RESULT_IGNORED, 1),
        # ffpuppet check ignored (log-limit)
        (False, FFPuppet.RC_WORKER, ["log-limit"], Target.RESULT_IGNORED, 1),
    ],
)
def test_puppet_target_03(mocker, tmp_path, healthy, reason, ignore, result, closes):
    """test PuppetTarget.detect_failure()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.RC_ALERT = FFPuppet.RC_ALERT
    fake_ffp.RC_CLOSED = FFPuppet.RC_CLOSED
    fake_ffp.RC_EXITED = FFPuppet.RC_EXITED
    fake_ffp.RC_WORKER = FFPuppet.RC_WORKER
    fake_file = tmp_path / "fake"
    fake_file.touch()
    target = PuppetTarget(str(fake_file), None, 300, 25, 5000)
    if "memory" in ignore:
        fake_ffp.return_value.available_logs.return_value = "ffp_worker_memory_usage"
    elif "log-limit" in ignore:
        fake_ffp.return_value.available_logs.return_value = "ffp_worker_log_size"
    fake_ffp.return_value.is_healthy.return_value = healthy
    fake_ffp.return_value.reason = reason
    assert target.detect_failure(ignore) == result
    assert fake_ffp.return_value.close.call_count == closes


@mark.parametrize(
    "healthy, usage, os_name, killed",
    [
        # skip sending SIGABRT on unsupported OSs
        (True, [(1234, 90)], "Windows", 0),
        # skip idle check if target is in a bad state
        (False, [], "Linux", 0),
        # send SIGABRT to hung process
        (True, [(234, 10), (236, 75), (238, 60)], "Linux", 1),
        # ignore idle timeout (close don't abort)
        (True, [(234, 10)], "Linux", 0),
    ],
)
def test_puppet_target_04(mocker, tmp_path, healthy, usage, os_name, killed):
    """test PuppetTarget.handle_hang()"""
    mocker.patch(
        "grizzly.target.puppet_target.system", autospec=True, return_value=os_name
    )
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_kill = mocker.patch("grizzly.target.puppet_target.kill", autospec=True)
    # raise OSError for code coverage
    fake_kill.side_effect = OSError
    fake_file = tmp_path / "fake"
    fake_file.touch()
    target = PuppetTarget(str(fake_file), None, 300, 25, 5000)
    fake_ffp.return_value.cpu_usage.return_value = usage
    fake_ffp.return_value.is_healthy.return_value = healthy
    target.handle_hang()
    assert fake_ffp.return_value.is_healthy.call_count == 1
    assert fake_ffp.return_value.close.call_count == 1
    assert fake_ffp.return_value.cpu_usage.call_count == (1 if usage else 0)
    assert fake_kill.call_count == fake_ffp.return_value.wait.call_count == killed


@mark.skipif(system() == "Windows", reason="Unsupported on Windows")
def test_puppet_target_05(mocker, tmp_path):
    """test PuppetTarget.dump_coverage()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_proc = mocker.patch("grizzly.target.puppet_target.Process", autospec=True)
    fake_proc.return_value.children.return_value = (mocker.Mock(pid=101),)
    fake_proc_iter = mocker.patch(
        "grizzly.target.puppet_target.process_iter", autospec=True
    )
    mocker.patch("grizzly.target.puppet_target.sleep", autospec=True)
    fake_time = mocker.patch("grizzly.target.puppet_target.time", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    target = PuppetTarget(str(fake_file), None, 300, 25, 5000)
    fake_kill = mocker.patch("grizzly.target.puppet_target.kill", autospec=True)
    # not running
    fake_ffp.return_value.get_pid.return_value = None
    target.dump_coverage()
    assert not fake_kill.call_count
    assert fake_ffp.return_value.get_pid.call_count == 1
    assert fake_proc_iter.call_count == 0
    # gcda not found
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.get_pid.return_value = 100
    fake_time.side_effect = (0, 1, 10)
    target.dump_coverage()
    assert fake_kill.call_count == 2
    assert fake_proc_iter.call_count == 2
    assert fake_ffp.return_value.is_healthy.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # browser crashes
    fake_ffp.return_value.is_healthy.side_effect = (True, False)
    fake_time.side_effect = None
    fake_time.return_value = 1.0
    target.dump_coverage()
    assert fake_kill.call_count == 2
    assert fake_proc_iter.call_count == 1
    assert fake_ffp.return_value.is_healthy.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # timeout while waiting for files
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.is_healthy.side_effect = None
    fake_ffp.return_value.get_pid.return_value = 100
    fake_proc_iter.return_value = (
        mocker.Mock(
            info={"pid": 100, "ppid": 0, "open_files": (mocker.Mock(path="a.gcda"),)}
        ),
    )
    fake_time.side_effect = (0, 1, 20, 20)
    target.dump_coverage(timeout=15)
    assert fake_kill.call_count == 3
    assert fake_proc_iter.call_count == 2
    assert fake_ffp.return_value.is_healthy.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # wait for files (success)
    fake_ffp.return_value.get_pid.return_value = 100
    fake_time.side_effect = None
    fake_time.return_value = 1.0
    fake_proc_iter.side_effect = (
        (
            mocker.Mock(
                info={
                    "pid": 100,
                    "ppid": 0,
                    "open_files": (
                        mocker.Mock(path="a.bin"),
                        mocker.Mock(path="/a/s/d"),
                    ),
                }
            ),
            mocker.Mock(info={"pid": 101, "ppid": 100, "open_files": None}),
            mocker.Mock(info={"pid": 999, "ppid": 0, "open_files": None}),
        ),
        (
            mocker.Mock(
                info={
                    "pid": 100,
                    "ppid": 0,
                    "open_files": (mocker.Mock(path="a.gcda"),),
                }
            ),
        ),
        (
            mocker.Mock(
                info={"pid": 100, "ppid": 0, "open_files": (mocker.Mock(path="a.bin"),)}
            ),
            mocker.Mock(
                info={
                    "pid": 999,
                    "ppid": 0,
                    "open_files": (mocker.Mock(path="ignore.gcda"),),
                }
            ),
        ),
    )
    target.dump_coverage()
    assert fake_proc_iter.call_count == 3
    assert fake_kill.call_count == 2
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()
    # kill calls raise OSError
    fake_kill.side_effect = OSError
    fake_ffp.return_value.is_healthy.return_value = True
    fake_ffp.return_value.get_pid.return_value = 100
    fake_proc_iter.side_effect = None
    fake_time.side_effect = count()
    target.dump_coverage()
    assert fake_kill.call_count == 3
    fake_ffp.reset_mock()
    fake_kill.reset_mock()
    fake_proc_iter.reset_mock()


def test_puppet_target_06(mocker, tmp_path):
    """test PuppetTarget.is_idle()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.return_value.cpu_usage.return_value = [(999, 30), (998, 20), (997, 10)]
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), None, 300, 25, 5000) as target:
        assert not target.is_idle(0)
        assert not target.is_idle(25)
        assert target.is_idle(50)


def test_puppet_target_07(mocker, tmp_path):
    """test PuppetTarget.monitor"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), None, 300, 25, 5000) as target:
        fake_ffp.return_value.is_running.return_value = False
        fake_ffp.return_value.is_healthy.return_value = False
        assert target.monitor is not None
        assert not target.monitor.is_healthy()
        assert not target.monitor.is_running()
        fake_ffp.return_value.is_running.return_value = True
        fake_ffp.return_value.is_healthy.return_value = True
        assert target.monitor.is_healthy()
        assert target.monitor.is_running()
        fake_ffp.return_value.launches = 123
        assert target.monitor.launches == 123
        fake_ffp.return_value.log_length.return_value = 100
        assert target.monitor.log_length("stdout") == 100
        target.monitor.clone_log("somelog")
        assert fake_ffp.return_value.clone_log.call_count == 1


def test_puppet_target_08(mocker, tmp_path):
    """test PuppetTarget.prefs"""
    mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(str(fake_file), None, 300, 25, 5000) as target:
        # default temp prefs
        assert not target._remove_prefs
        prefs_file = target.prefs
        assert isfile(prefs_file)
        assert target._remove_prefs
        # set prefs, remove temp
        target.prefs = str(fake_file)
        assert not isfile(prefs_file)
        assert not target._remove_prefs
        prefs_file = target.prefs
        assert isfile(prefs_file)
        # unset prefs (revert to temp), don't remove previously specified file
        target.prefs = None
        assert target._remove_prefs
        assert isfile(prefs_file)
        prefs_file = target.prefs
        # set missing file
        with raises(TargetError, match="Missing prefs.js file 'missing'"):
            target.prefs = "missing"
        assert not isfile(prefs_file)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from abc import ABCMeta, abstractmethod
from collections import namedtuple
from hashlib import sha1
from json import dump, dumps, loads
from logging import WARNING, getLogger
from os import SEEK_END, getcwd, getenv, makedirs, mkdir, scandir, stat, unlink, walk
from os.path import basename, expanduser, isdir, isfile
from os.path import join as pathjoin
from os.path import realpath, relpath
from platform import machine, system
from re import DOTALL, VERBOSE
from re import compile as re_compile
from shutil import copyfile, copyfileobj, move, rmtree
from tarfile import open as tar_open
from tempfile import mkstemp
from time import strftime
from zipfile import ZIP_DEFLATED, ZipFile

# import FuzzManager utilities
from Collector.Collector import Collector
from fasteners.process_lock import InterProcessLock
from FTB.ProgramConfiguration import ProgramConfiguration
from FTB.Signatures.CrashInfo import CrashInfo
from psutil import disk_usage

# check if boto is available for S3FuzzManager reporter
try:
    from boto3 import resource
    from botocore.exceptions import ClientError

    _boto_import_error = None  # pylint: disable=invalid-name
    getLogger("botocore").setLevel(WARNING)
    getLogger("boto3").setLevel(WARNING)
    getLogger("s3transfer").setLevel(WARNING)
except ImportError as err:
    _boto_import_error = err  # pylint: disable=invalid-name

from .stack_hasher import Stack
from .utils import grz_tmp

__all__ = (
    "FilesystemReporter",
    "FuzzManagerReporter",
    "Report",
    "S3FuzzManagerReporter",
)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)

# NOTE: order matters, aux -> stderr -> stdout
LogMap = namedtuple("LogMap", "aux stderr stdout")


class Report:
    DEFAULT_MAJOR = "NO_STACK"
    DEFAULT_MINOR = "0"
    HANG_STACK_HEIGHT = 10
    MAX_LOG_SIZE = 1048576  # 1MB

    __slots__ = (
        "_crash_info",
        "_logs",
        "_signature",
        "_target_binary",
        "is_hang",
        "path",
        "prefix",
        "stack",
    )

    def __init__(self, log_path, target_binary, is_hang=False, size_limit=MAX_LOG_SIZE):
        assert isinstance(log_path, str)
        assert isinstance(target_binary, str)
        self._crash_info = None
        self._logs = self.select_logs(log_path)
        assert self._logs is not None
        self._signature = None
        self._target_binary = target_binary
        self.is_hang = is_hang
        self.path = log_path
        # tail files in log_path if needed
        if size_limit < 1:
            LOG.warning("No limit set on report log size!")
        else:
            with scandir(path=log_path) as contents:
                for log in contents:
                    if log.is_file() and log.stat().st_size > size_limit:
                        Report.tail(log.path, size_limit)
        # look through logs one by one until we find a stack
        for log_file in (x for x in self._logs if x is not None):
            with open(log_file, "rb") as log_fp:
                stack = Stack.from_text(log_fp.read().decode("utf-8", errors="ignore"))
            if stack.frames:
                # limit the hash calculations to the first n frames if a hang
                # was detected to attempt to help local bucketing
                stack.height_limit = self.HANG_STACK_HEIGHT if is_hang else None
                self.prefix = "%s_%s" % (stack.minor[:8], strftime("%Y-%m-%d_%H-%M-%S"))
                self.stack = stack
                break
        else:
            self.prefix = "%s_%s" % (self.DEFAULT_MINOR, strftime("%Y-%m-%d_%H-%M-%S"))
            self.stack = None

    @staticmethod
    def calc_hash(signature):
        """Create unique hash from a signature.

        Args:
            None

        Returns:
            str: Hash of the raw signature.
        """
        return sha1(signature.rawSignature.encode("utf-8")).hexdigest()[:16]

    def cleanup(self):
        """Remove Report data from filesystem.

        Args:
            None

        Returns:
            None
        """
        if self.path and isdir(self.path):
            rmtree(self.path)
        self.path = None

    @property
    def crash_hash(self):
        """Create unique hash from a signature.

        Args:
            None

        Returns:
            str: Hash of the raw signature of the crash.
        """
        return self.calc_hash(self.crash_signature)

    @property
    def crash_info(self):
        """Create CrashInfo object from logs.

        Args:
            None

        Returns:
            CrashInfo: CrashInfo based on log data.
        """
        if self._crash_info is None:
            assert self.path is not None
            # read in the log files and create a CrashInfo object
            if self._logs.aux is not None:
                with open(self._logs.aux, "rb") as log_fp:
                    aux_data = (
                        log_fp.read().decode("utf-8", errors="ignore").splitlines()
                    )
            else:
                aux_data = None
            # create ProgramConfiguration that can be reported to a FM server
            if isfile("%s.fuzzmanagerconf" % (self._target_binary,)):
                # attempt to use "<target_binary>.fuzzmanagerconf"
                fm_cfg = ProgramConfiguration.fromBinary(self._target_binary)
            else:
                LOG.debug("'%s.fuzzmanagerconf' does not exist", self._target_binary)
                fm_cfg = None
            if fm_cfg is None:
                LOG.debug("creating ProgramConfiguration")
                cpu = machine().lower()
                fm_cfg = ProgramConfiguration(
                    basename(self._target_binary),
                    "x86_64" if cpu == "amd64" else cpu,
                    system(),
                )
            with open(self._logs.stderr, "rb") as err_fp, open(
                self._logs.stdout, "rb"
            ) as out_fp:
                self._crash_info = CrashInfo.fromRawCrashData(
                    out_fp.read().decode("utf-8", errors="ignore").splitlines(),
                    err_fp.read().decode("utf-8", errors="ignore").splitlines(),
                    fm_cfg,
                    auxCrashData=aux_data,
                )
        return self._crash_info

    @property
    def crash_signature(self):
        """Create CrashSignature object from CrashInfo.

        Args:
            None

        Returns:
            CrashSignature: CrashSignature based on log data.
        """
        if self._signature is None:
            self._signature = self.crash_info.createCrashSignature(
                maxFrames=self.crash_signature_max_frames(self.crash_info)
            )
        return self._signature

    @staticmethod
    def crash_signature_max_frames(crash_info, suggested_frames=8):
        if set(crash_info.backtrace) & {
            "std::panicking::rust_panic",
            "std::panicking::rust_panic_with_hook",
        }:
            # rust panic adds 5-6 frames of noise at the top of the stack
            suggested_frames += 6
        return suggested_frames

    @staticmethod
    def _find_ffpuppet_worker(logs):
        """Search through list of log files for a ffpuppet worker log.

        Args:
            logs (list(str)): List of log files to search.

        Returns:
            str: The full file path if a match is found otherwise None.
        """
        found = None
        for fname in (x for x in logs if "ffp_worker" in x):
            if found is not None:
                # we only expect one log here...
                LOG.warning("overwriting previously selected %r", found)
            found = fname
        return found

    @staticmethod
    def _find_minidump(logs):
        """Search through list of log files for a minidump log.

        Args:
            logs (list(str)): List of log files to search.

        Returns:
            str: The full file path if a match is found otherwise None.
        """
        re_dump_req = re_compile(
            r"\d+\|0\|.+?\|google_breakpad::ExceptionHandler::WriteMinidump"
        )
        for fname in (x for x in logs if "minidump" in x):
            with open(fname, "r") as log_fp:
                data = log_fp.read(65536)
                # this will select log that contains "Crash|SIGSEGV|" or
                # the desired "DUMP_REQUESTED" log
                # TODO: review this it may be too strict
                # see mozilla-central/source/accessible/ipc/DocAccessibleParent.cpp#452
                if "Crash|DUMP_REQUESTED|" not in data or re_dump_req.search(data):
                    return fname
        return None

    @staticmethod
    def _find_sanitizer(logs):
        """Search through list of log files for a sanitizer (ASan, UBSan, etc...) log.

        Args:
            logs (list(str)): List of log files to search.

        Returns:
            str: The full file path if a match is found otherwise None.
        """
        # pattern to identify the ASan crash triggered when the parent process goes away
        # TODO: this may no longer be required
        re_e10s_forced = re_compile(
            r"""
            ==\d+==ERROR:.+?SEGV\son.+?0x[0]+\s\(.+?T2\).+?
            #0\s+0x[0-9a-f]+\sin\s+mozilla::ipc::MessageChannel::OnChannelErrorFromLink
            """,
            DOTALL | VERBOSE,
        )
        # this is a list of Sanitizer error reports to prioritize
        # Sanitizer reports not included below are deprioritized
        prioritize_tokens = (
            "use-after-",
            "-buffer-overflow on",
            ": SEGV on ",
            "access-violation on ",
            "attempting free on ",
            "negative-size-param",
            "-param-overlap",
        )
        found = None
        for fname in (x for x in logs if "asan" in x):
            with open(fname, "r") as log_fp:
                data = log_fp.read(65536)
            # look for interesting crash info in the log
            if "==ERROR:" in data:
                # check for e10s forced crash
                if re_e10s_forced.search(data) is not None:
                    continue
                # make sure there is something that looks like a stack frame
                if "#0 " in data:
                    found = fname
                    if any(x in data for x in prioritize_tokens):
                        # this is the likely cause of the crash
                        break
            if found is None:
                # UBSan error (non-ASan builds)
                if ": runtime error: " in data:
                    found = fname
                # catch all (choose the one with info for now)
                elif data:
                    found = fname
        return found

    @staticmethod
    def _find_valgrind(logs):
        """Search through list of log files for a Valgrind worker log.

        Args:
            logs (list(str)): List of log files to search.

        Returns:
            str: The full file path if a match is found otherwise None.
        """
        for fname in (x for x in logs if "valgrind" in x):
            if stat(fname).st_size:
                return fname
        return None

    @property
    def major(self):
        """The inclusive bucketing hash based on the stack trace
        data found in logs.

        Args:
            None

        Returns:
            str: major hash string.
        """
        if self.stack and self.stack.major is not None:
            return self.stack.major
        return self.DEFAULT_MAJOR

    @property
    def minor(self):
        """The specific bucketing hash based on the stack trace
        data found in logs.

        Args:
            None

        Returns:
            str: minor hash string.
        """
        if self.stack and self.stack.minor is not None:
            return self.stack.minor
        return self.DEFAULT_MINOR

    @property
    def preferred(self):
        """Log file containing most relevant data.

        Args:
            None

        Returns:
            str: Name of log.
        """
        return self._logs.aux or self._logs.stderr

    @classmethod
    def select_logs(cls, path):
        """Scan path for file containing stderr, stdout and other (aux)
        data and build a LogMap.

        Args:
            path (str): Path to scan for log files.

        Returns:
            LogMap: A LogMap pointing to log files or None if path is empty.
        """
        to_scan = None
        with scandir(path=path) as contents:
            files = (x for x in contents if x.is_file())
            # order by date hopefully the oldest log is the cause of the issue
            to_scan = [x.path for x in sorted(files, key=lambda x: x.stat().st_mtime)]
        if not to_scan:
            LOG.warning("No files found in %r", path)
            return None
        # look for file to use as aux log
        log_aux = cls._find_sanitizer(to_scan)
        if log_aux is None:
            log_aux = cls._find_valgrind(to_scan)
        if log_aux is None:
            log_aux = cls._find_minidump(to_scan)
        if log_aux is None:
            log_aux = cls._find_ffpuppet_worker(to_scan)
        # look for stderr and stdout log files
        log_err = None
        log_out = None
        for fname in to_scan:
            if "stderr" in fname:
                log_err = fname
            elif "stdout" in fname:
                log_out = fname
        result = LogMap(log_aux, log_err, log_out)
        return result if any(result) else None

    @staticmethod
    def tail(in_file, size_limit):
        """Tail the given file. WARNING: This is destructive!

        Args:
            in_file (str): Path to file to work with.
            size_limit (int): Maximum size of file after tail operation.

        Returns:
            None
        """
        assert size_limit > 0
        with open(in_file, "rb") as in_fp:
            in_fp.seek(0, SEEK_END)
            end = in_fp.tell()
            if end <= size_limit:
                return
            dump_pos = end - size_limit
            in_fp.seek(dump_pos)
            out_fd, out_file = mkstemp(prefix="taillog_", dir=grz_tmp())
            with open(out_fd, "wb") as out_fp:
                out_fp.write(b"[LOG TAILED]\n")
                copyfileobj(in_fp, out_fp, 0x10000)  # 64KB chunks
        unlink(in_file)
        move(out_file, in_file)


class Reporter(metaclass=ABCMeta):
    @abstractmethod
    def _post_submit(self):
        pass

    @abstractmethod
    def _pre_submit(self, report):
        pass

    @abstractmethod
    def _submit_report(self, report, test_cases):
        pass

    def submit(self, test_cases, report):
        """Submit report containing results.

        Args:
            test_cases (iterable): A collection of testcases, ordered newest to oldest,
                                   the newest being the mostly likely to trigger
                                   the result (crash, assert... etc).
            report (Report): Report to submit.

        Returns:
            *: implementation specific result indicating where the report was created
        """
        assert isinstance(report, Report)
        assert report.path is not None
        self._pre_submit(report)
        result = self._submit_report(report, test_cases)
        if report is not None:
            report.cleanup()
        self._post_submit()
        return result


class FilesystemReporter(Reporter):
    DISK_SPACE_ABORT = 512 * 1024 * 1024  # 512 MB

    __slots__ = ("major_bucket", "report_path")

    def __init__(self, report_path, major_bucket=True):
        self.major_bucket = major_bucket
        self.min_space = FilesystemReporter.DISK_SPACE_ABORT
        assert isinstance(report_path, str) and report_path
        self.report_path = report_path

    def _pre_submit(self, report):
        pass

    def _post_submit(self):
        pass

    def _submit_report(self, report, test_cases):
        # create major bucket directory in working directory if needed
        if self.major_bucket:
            dest_path = pathjoin(self.report_path, report.major[:16])
        else:
            dest_path = self.report_path
        if not isdir(dest_path):
            makedirs(dest_path)
        # dump test cases and the contained files to working directory
        for test_number, test_case in enumerate(test_cases):
            dump_path = pathjoin(dest_path, "%s-%d" % (report.prefix, test_number))
            if not isdir(dump_path):
                mkdir(dump_path)
            test_case.dump(dump_path, include_details=True)
        # move logs into bucket directory
        log_path = pathjoin(dest_path, "%s_%s" % (report.prefix, "logs"))
        if isdir(log_path):
            LOG.warning("Report log path exists %r", log_path)
        move(report.path, log_path)
        # avoid filling the disk
        free_space = disk_usage(log_path).free
        if free_space < self.min_space:
            raise RuntimeError(
                "Running low on disk space (%0.1fMB)" % (free_space / 1048576.0,)
            )
        return dest_path


class FuzzManagerReporter(Reporter):
    FM_CONFIG = pathjoin(expanduser("~"), ".fuzzmanagerconf")
    # max number of times to report a non-frequent signature to FuzzManager
    MAX_REPORTS = 10

    # testcase quality values
    QUAL_REDUCED_RESULT = 0  # the final reduced testcase
    QUAL_REDUCED_ORIGINAL = 1  # the original used for successful reduction
    QUAL_REDUCING = 4  # the testcase is currently being reduced
    QUAL_UNREDUCED = 5  # haven't attempted reduction yet (1st attempt, generic reducer)
    QUAL_REQUEST_SPECIFIC = 6  # platform specific reduction requested (2nd attempt)
    QUAL_NO_TESTCASE = 7  # testcase not detected ("testcase" not a testcase?)
    QUAL_REDUCER_BROKE = 8  # the testcase was reproducible, but broke during reduction
    QUAL_REDUCER_ERROR = 9  # reducer error
    QUAL_NOT_REPRODUCIBLE = 10  # could not reproduce the testcase

    def __init__(self, tool=None):
        self._extra_metadata = {}
        self.force_report = False
        self.max_reports = FuzzManagerReporter.MAX_REPORTS
        self.quality = self.QUAL_UNREDUCED
        self.tool = tool  # optional tool name

    def _post_submit(self):
        self._extra_metadata.clear()

    @staticmethod
    def sanity_check(bin_file):
        """Perform FuzzManager sanity check.

        Args:
            bin_file (str): Binary file being tested.

        Returns:
            None
        """
        if not isfile(FuzzManagerReporter.FM_CONFIG):
            raise IOError("Missing: %s" % (FuzzManagerReporter.FM_CONFIG,))
        if not isfile("".join([bin_file, ".fuzzmanagerconf"])):
            raise IOError("Missing: %s.fuzzmanagerconf" % (bin_file,))
        ProgramConfiguration.fromBinary(bin_file)

    def add_extra_metadata(self, key, value):
        """Add extra metadata to be reported with any CrashEntrys reported.

        Arguments:
            key (str): key for this data in the metadata dict
            value (object): JSON serializable object to be included in the FM crash
                            metadata. The object will be deep-copied.

        Returns:
            None
        """
        assert isinstance(key, str)
        assert key not in self._extra_metadata
        # deep copy and ensure that value is JSON serializable
        self._extra_metadata[key] = loads(dumps(value))

    @classmethod
    def quality_name(cls, value):
        for name in dir(cls):
            if name.startswith("QUAL_") and getattr(cls, name) == value:
                return name
        return "unknown quality (%r)" % (value,)

    def _pre_submit(self, report):
        self._process_rr_trace(report)

    def _process_rr_trace(self, report):
        # don't report large files to FuzzManager
        trace_path = pathjoin(report.path, "rr-traces")
        if isdir(trace_path):
            LOG.info("Ignored rr trace")
            self.add_extra_metadata("rr-trace", "ignored")
            # remove traces so they are not uploaded to FM (because they are huge)
            # use S3FuzzManagerReporter instead
            rmtree(trace_path)

    @staticmethod
    def _ignored(report):
        # This is here to prevent reporting stack-less crashes
        # that were caused by system OOM or bogus other crashes
        with open(report.preferred, "rb") as log_fp:
            log_data = log_fp.read().decode("utf-8", errors="ignore")
        mem_errs = ("ERROR: Failed to mmap", ": AddressSanitizer failed to allocate")
        # ignore sanitizer OOMs missing stack
        for msg in mem_errs:
            if msg in log_data and "#0 " not in log_data:
                return True
        # ignore Valgrind crashes
        if log_data.startswith("VEX temporary storage exhausted."):
            return True
        return False

    def _submit_report(self, report, test_cases):
        # search for a cached signature match and if the signature
        # is already in the cache and marked as frequent, don't bother submitting
        with InterProcessLock(pathjoin(grz_tmp(), "fm_sigcache.lock")):
            collector = Collector()
            cache_sig_file, cache_metadata = collector.search(report.crash_info)
            if cache_metadata is not None:
                if cache_metadata["frequent"]:
                    LOG.info(
                        "Frequent crash matched existing signature: %s",
                        cache_metadata["shortDescription"],
                    )
                    if not self.force_report:
                        return None
                elif "bug__id" in cache_metadata:
                    LOG.info(
                        "Crash matched existing signature (bug %s): %s",
                        cache_metadata["bug__id"],
                        cache_metadata["shortDescription"],
                    )
                    # we will still report this one, but no more
                    cache_metadata["frequent"] = True
                # there is already a signature, initialize count
                cache_metadata.setdefault("_grizzly_seen_count", 0)
            else:
                # there is no signature, create one locally so we can count
                # the number of times we've seen it
                max_frames = report.crash_signature_max_frames(report.crash_info)
                cache_sig_file = collector.generate(
                    report.crash_info, numFrames=max_frames
                )
                cache_metadata = {
                    "_grizzly_seen_count": 0,
                    "frequent": False,
                    "shortDescription": report.crash_info.createShortSignature(),
                }
            if cache_sig_file is None:
                if self._ignored(report):
                    LOG.info("Report is unsupported and is in ignore list")
                    return None
                LOG.warning("Report is unsupported by FM, saved to %r", report.path)
                # TODO: we should check if stackhasher failed too
                raise RuntimeError("Failed to create FM signature")
            # limit the number of times we report per cycle
            cache_metadata["_grizzly_seen_count"] += 1
            if cache_metadata["_grizzly_seen_count"] >= self.max_reports:
                # we will still report this one, but no more
                cache_metadata["frequent"] = True
            metadata_file = cache_sig_file.replace(".signature", ".metadata")
            with open(metadata_file, "w") as meta_fp:
                dump(cache_metadata, meta_fp)

        if report.is_hang:
            self.add_extra_metadata("is_hang", True)

        # dump test cases and the contained files to working directory
        test_case_meta = []
        for test_number, test_case in enumerate(test_cases):
            test_case_meta.append([test_case.adapter_name, test_case.input_fname])
            dump_path = pathjoin(report.path, "%s-%d" % (report.prefix, test_number))
            if not isdir(dump_path):
                mkdir(dump_path)
            test_case.dump(dump_path, include_details=True)
        report.crash_info.configuration.addMetadata(
            {"grizzly_input": repr(test_case_meta)}
        )
        if test_cases:
            environ_string = " ".join(
                "=".join(kv) for kv in test_cases[0].env_vars.items()
            )
            report.crash_info.configuration.addMetadata(
                {"recorded_envvars": environ_string}
            )
        else:
            self.quality = self.QUAL_NO_TESTCASE
        report.crash_info.configuration.addMetadata(self._extra_metadata)

        # grab screen log (used in automation)
        if getenv("WINDOW") is not None:
            screen_log = pathjoin(getcwd(), ".".join(["screenlog", getenv("WINDOW")]))
            if isfile(screen_log):
                target_log = pathjoin(report.path, "screenlog.txt")
                copyfile(screen_log, target_log)
                Report.tail(target_log, 10240)  # limit to last 10K

        # add results to a zip file
        zip_name = "%s.zip" % (report.prefix,)
        with ZipFile(zip_name, mode="w", compression=ZIP_DEFLATED) as zip_fp:
            # add test files
            for dir_name, _, dir_files in walk(report.path):
                arc_path = relpath(dir_name, report.path)
                for file_name in dir_files:
                    zip_fp.write(
                        pathjoin(dir_name, file_name),
                        arcname=pathjoin(arc_path, file_name),
                    )

        # override tool name if specified
        if self.tool is not None:
            collector.tool = self.tool

        # announce shortDescription if crash is not in a bucket
        if (
            cache_metadata["_grizzly_seen_count"] == 1
            and not cache_metadata["frequent"]
        ):
            LOG.info("Submitting new crash %r", cache_metadata["shortDescription"])
        # submit results to the FuzzManager server
        new_entry = collector.submit(
            report.crash_info, testCase=zip_name, testCaseQuality=self.quality
        )
        LOG.info("Logged %d with quality %d", new_entry["id"], self.quality)

        # remove zipfile
        if isfile(zip_name):
            unlink(zip_name)

        return new_entry["id"]


class S3FuzzManagerReporter(FuzzManagerReporter):
    @staticmethod
    def compress_rr_trace(src, dest):
        # resolve symlink to latest trace available
        latest_trace = realpath(pathjoin(src, "latest-trace"))
        assert isdir(latest_trace), "missing latest-trace directory"
        rr_arc = pathjoin(dest, "rr.tar.bz2")
        LOG.debug("creating %r from %r", rr_arc, latest_trace)
        with tar_open(rr_arc, "w:bz2") as arc_fp:
            arc_fp.add(latest_trace, arcname=basename(latest_trace))
        # remove path containing uncompressed traces
        rmtree(src)
        return rr_arc

    def _pre_submit(self, report):
        self._process_rr_trace(report)

    def _process_rr_trace(self, report):
        trace_path = pathjoin(report.path, "rr-traces")
        if not isdir(trace_path):
            return None
        s3_bucket = getenv("GRZ_S3_BUCKET")
        assert s3_bucket is not None
        # check for existing minor hash in S3
        s3_res = resource("s3")
        s3_key = "rr-%s.tar.bz2" % (report.minor,)
        s3_url = "http://%s.s3.amazonaws.com/%s" % (s3_bucket, s3_key)
        try:
            # HEAD, doesn't fetch the whole object
            s3_res.Object(s3_bucket, s3_key).load()
        except ClientError as exc:
            if exc.response["Error"]["Code"] == "404":
                # The object does not exist.
                pass
            else:  # pragma: no cover
                # Something else has gone wrong.
                raise
        else:
            # The object already exists.
            LOG.info("rr trace exists at %r", s3_url)
            self.add_extra_metadata("rr-trace", s3_url)
            # remove traces so they are not reported to FM
            rmtree(trace_path)
            return s3_url

        # Upload to S3
        rr_arc = self.compress_rr_trace(trace_path, report.path)
        s3_res.meta.client.upload_file(
            rr_arc, s3_bucket, s3_key, ExtraArgs={"ACL": "public-read"}
        )
        unlink(rr_arc)
        self.add_extra_metadata("rr-trace", s3_url)
        return s3_url

    @staticmethod
    def sanity_check(bin_file):
        if getenv("GRZ_S3_BUCKET") is None:
            raise EnvironmentError("'GRZ_S3_BUCKET' is not set in environment")
        if _boto_import_error is not None:
            raise _boto_import_error  # pylint: disable=raising-bad-type
        FuzzManagerReporter.sanity_check(bin_file)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Manage Grizzly status reports."""
from collections import defaultdict, namedtuple
from contextlib import contextmanager
from json import dump, load
from logging import getLogger
from os import close, getpid, listdir, unlink
from os.path import isdir, isfile
from os.path import join as pathjoin
from tempfile import mkstemp
from time import time

from fasteners.process_lock import InterProcessLock

from .utils import grz_tmp

__all__ = ("Status",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)

ProfileEntry = namedtuple("ProfileEntry", "count max min name total")


class Status:
    """Status holds status information for the Grizzly session.
    There can be multiple readers of the data but only a single writer.
    """

    PATH = grz_tmp("status")
    REPORT_FREQ = 60

    __slots__ = (
        "_enable_profiling",
        "_lock",
        "_profiles",
        "_results",
        "data_file",
        "ignored",
        "iteration",
        "log_size",
        "pid",
        "start_time",
        "test_name",
        "timestamp",
    )

    def __init__(self, data_file, enable_profiling=False, start_time=None):
        assert data_file.endswith(".json")
        assert start_time is None or isinstance(start_time, float)
        self._lock = InterProcessLock("%s.lock" % (data_file,))
        self._enable_profiling = enable_profiling
        self._profiles = dict()
        self._results = defaultdict(int)
        # if data_file is None the status report is read only (no reporting)
        self.data_file = data_file
        self.ignored = 0
        self.iteration = 0
        self.log_size = 0
        self.pid = None
        self.start_time = start_time
        self.test_name = None
        self.timestamp = start_time

    def cleanup(self):
        """Remove data and lock files from disk.

        Args:
            None

        Returns:
            None
        """
        if self.data_file is None:
            return
        try:
            with self._lock:
                unlink(self.data_file)
        except OSError:  # pragma: no cover
            LOG.warning("Failed to delete %r", self.data_file)
        try:
            unlink("%s.lock" % (self.data_file,))
        except OSError:  # pragma: no cover
            pass
        self.data_file = None

    def count_result(self, signature):
        """Increment counter that matches `signature`.

        Args:
            signature (str):

        Returns:
            None
        """
        self._results[signature] += 1

    @property
    def _data(self):
        return {
            "_profiles": self._profiles,
            "_results": self._results,
            "ignored": self.ignored,
            "iteration": self.iteration,
            "log_size": self.log_size,
            "pid": self.pid,
            "start_time": self.start_time,
            "test_name": self.test_name,
            "timestamp": self.timestamp,
        }

    @property
    def duration(self):
        """Calculate the number of seconds since start() was called.

        Args:
            None

        Returns:
            int: Total runtime in seconds since start() was called
        """
        return max(self.timestamp - self.start_time, 0)

    @classmethod
    def load(cls, data_file):
        """Load status report. Loading a status report from disk will create a
        read only status report.

        Args:
            data_file (str): JSON file that contains status data.

        Returns:
            Status: Loaded status object or None
        """
        status = cls(data_file)
        data = None
        try:
            with status._lock:  # pylint: disable=protected-access
                with open(data_file, "r") as out_fp:
                    data = load(out_fp)
        except OSError:
            LOG.debug("failed to open %r", data_file)
            # if data_file exists the lock will be removed by the active session
            if not isfile(data_file):
                # attempt to remove potentially leaked lock file
                try:
                    unlink("%s.lock" % (data_file,))
                except OSError:  # pragma: no cover
                    pass
        except ValueError:
            LOG.debug("failed to load json data from %r", data_file)
        else:
            LOG.debug("no such file %r", data_file)
        if data is None:
            return None
        if "start_time" not in data:
            LOG.debug("invalid status json file")
            return None
        for attr, value in data.items():
            setattr(status, attr, value)
        # set read only
        status.data_file = None
        return status

    @classmethod
    def loadall(cls):
        """Load all status reports found in cls.PATH.

        Args:
            None

        Returns:
            Generator: Status objects stored in cls.PATH.
        """
        if isdir(cls.PATH):
            for data_file in listdir(cls.PATH):
                if not data_file.endswith(".json"):
                    continue
                status = cls.load(pathjoin(cls.PATH, data_file))
                if status is None:
                    continue
                yield status

    @contextmanager
    def measure(self, name):
        """Used to simplify collecting profiling data.

        Args:
            name (str): Used to group the entries.

        Returns:
            None
        """
        if self._enable_profiling:
            mark = time()
            yield
            self.record(name, time() - mark)
        else:
            yield

    def profile_entries(self):
        """Used to retrieve profiling data.

        Args:
            None

        Yields:
            ProfileEntry: Containing recorded profiling data.
        """
        for name, entry in self._profiles.items():
            yield ProfileEntry(
                entry["count"],
                entry["max"],
                entry["min"],
                name,
                entry["total"],
            )

    @property
    def rate(self):
        """Calculate the number of iterations performed per second since start()
        was called.

        Args:
            None

        Returns:
            float: Number of iterations performed per second.
        """
        return self.iteration / float(self.duration) if self.duration > 0 else 0

    def record(self, name, duration):
        """Used to add profiling data. This is intended to be used to make rough
        calculations to identify major configuration issues.

        Args:
            name (str): Used to group the entries.
            duration (int, float): Stored to be later used for measurements.

        Returns:
            None
        """
        assert isinstance(duration, (float, int))
        try:
            self._profiles[name]["count"] += 1
            if self._profiles[name]["max"] < duration:
                self._profiles[name]["max"] = duration
            elif self._profiles[name]["min"] > duration:
                self._profiles[name]["min"] = duration
            self._profiles[name]["total"] += duration
        except KeyError:
            if self._enable_profiling:
                # add profile entry
                self._profiles[name] = {
                    "count": 1,
                    "max": duration,
                    "min": duration,
                    "total": duration,
                }

    def report(self, force=False, report_freq=REPORT_FREQ):
        """Write status report to disk. Reports are only written periodically.
        It is limited by `report_freq`. The specified number of seconds must
        elapse before another write will be performed unless `force` is True.

        Args:
            force (bool): Ignore report frequently limiting.
            report_freq (int): Minimum number of seconds between writes.

        Returns:
            bool: Returns true if the report was successful otherwise false.
        """
        assert self.data_file is not None
        now = time()
        if not force and now < (self.timestamp + report_freq):
            return False
        self.timestamp = now
        with self._lock:
            with open(self.data_file, "w") as out_fp:
                dump(self._data, out_fp)
        return True

    @property
    def results(self):
        """Calculate the total number of results.

        Args:
            None

        Returns:
            int: Total number of results.
        """
        return sum(self._results.values())

    def signatures(self):
        """Provide the signature and the number of times it has been found for
        each result.

        Args:
            None

        Yields:
            tuples: Containing signature and count.
        """
        for sig, count in self._results.items():
            yield (sig, count)

    @classmethod
    def start(cls, enable_profiling=False):
        """Create a unique Status object.

        Args:
            enable_profiling (bool): Record profiling data.

        Returns:
            Status: Active status report.
        """
        tfd, filepath = mkstemp(dir=cls.PATH, prefix="grzstatus_", suffix=".json")
        close(tfd)
        status = cls(filepath, enable_profiling=enable_profiling, start_time=time())
        status.pid = getpid()
        status.report(force=True)
        return status



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Grizzly Reduction exceptions."""
from ..session import Session


class GrizzlyReduceBaseException(Exception):
    """Base for other Grizzly Reducer specific exceptions."""

    def __init__(self, msg, code=Session.EXIT_ERROR):
        super().__init__()
        self.msg = msg
        self.code = code


class NotReproducible(GrizzlyReduceBaseException):
    """Crash was not observed when expected during reduction."""

    def __init__(self, msg):
        super().__init__(msg, code=Session.EXIT_FAILURE)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""CLI argument parsing for Grizzly reduction.
"""
from logging import getLogger
from pathlib import Path

from ..replay.args import ReplayArgs
from .strategies import DEFAULT_STRATEGIES, STRATEGIES

LOG = getLogger(__name__)


class ReduceArgs(ReplayArgs):
    """Argument parser for `grizzly.reduce`.

    Takes all arguments defined for `grizzly.replay`, and a few specific to reduction.
    """

    def __init__(self):
        """Initialize argument parser."""
        super().__init__()

        # these arguments have other defaults vs how they are defined in ReplayArgs
        self.parser.set_defaults(
            include_test=True,
            logs=".",
        )

        reduce_args = self.parser.add_argument_group("Reduce Arguments")
        reduce_args.add_argument(
            "--no-analysis",
            action="store_true",
            help="Disable analysis to auto-set --repeat/--min-crashes.",
        )
        reduce_args.add_argument(
            "--report-period",
            type=int,
            help="Periodically report the best testcase for long-running strategies."
            " (value in seconds, default: no)",
        )
        reduce_args.add_argument(
            "--static-timeout",
            action="store_true",
            help="Disable automatically updating the iteration timeout.",
        )
        reduce_args.add_argument(
            "--strategy",
            nargs="+",
            default=DEFAULT_STRATEGIES,
            metavar="STRATEGY",
            dest="strategies",
            help="One or more strategies (space-separated). Available: %s (default: %s)"
            % (" ".join(sorted(STRATEGIES)), " ".join(DEFAULT_STRATEGIES)),
        )

    def sanity_check(self, args):
        """Sanity check reducer args.

        Arguments:
            args (argparse.Namespace): Result from `parse_args()`.

        Raises:
            SystemExit: on error, `ArgumentParser.error()` is called, which will exit.
        """
        super().sanity_check(args)

        # if logs is specified, we need it to be a directory (whether existent or not)
        if Path(args.logs).is_file():
            self.parser.error("'--logs' cannot be a file")

        # check that specified strategies exist
        for strategy in args.strategies:
            if strategy not in STRATEGIES:
                self.parser.error("Unrecognized '--strategy': '%s'" % (strategy,))

        if args.report_period is not None:
            if args.report_period <= 0:
                self.parser.error("Invalid --report-period (value is in seconds)")
            if args.report_period < 60:
                self.parser.error("Very short --report-period (value is in seconds)")

        if not args.no_analysis:
            # analysis is enabled, but repeat/min_crashes specified. doesn't make sense
            errors = []
            if args.repeat != self.parser.get_default("repeat"):
                errors.append("'--repeat'")
            if args.min_crashes != self.parser.get_default("min_crashes"):
                errors.append("'--min-crashes'")
            if errors:
                error_str = " and ".join(errors)
                LOG.warning(
                    "%s specified, with analysis enabled, they will be ignored",
                    error_str,
                )


class ReduceFuzzManagerIDArgs(ReduceArgs):
    def __init__(self):
        """Initialize argument parser."""
        super().__init__()

        # madhax alert!
        #
        # We need to modify the meaning of the 'input' positional to accept an int ID
        # instead of a local testcase.
        # This is not possible with the public argparse API.
        #
        # refs: stackoverflow.com/questions/32807319/disable-remove-argument-in-argparse
        #       bugs.python.org/issue19462

        # look up the action for the positional `input` arg
        action = None
        for arg in self.parser._actions:
            if arg.dest == "input" and not arg.option_strings:
                action = arg
                break
        assert action is not None

        # modify it's type and help string
        action.type = int
        action.help = "FuzzManager ID to reduce"

        # ... and Bob's your uncle
        self._sanity_skip.add("input")


class ReduceFuzzManagerIDQualityArgs(ReduceFuzzManagerIDArgs):
    def __init__(self):
        """Initialize argument parser."""
        super().__init__()
        self.parser.add_argument(
            "--quality", type=int, help="Only try crashes with a given quality value"
        )

    def sanity_check(self, args):
        super().sanity_check(args)

        if args.quality is not None and args.quality < 0:
            self.parser.error("'--quality' value cannot be negative")



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .args import GrizzlyArgs
from .main import main

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


raise SystemExit(main(GrizzlyArgs().parse_args()))



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger

from ..common.fuzzmanager import Bucket, CrashEntry
from ..common.reporter import FuzzManagerReporter
from ..main import configure_logging
from ..session import Session
from .args import ReduceFuzzManagerIDArgs
from .core import ReduceManager

LOG = getLogger(__name__)


def main(args):
    """CLI for `grizzly.reduce.crash`.

    Arguments:
        args (argparse.Namespace): Result from `ReduceArgs.parse_args`.

    Returns:
        int: 0 for success. non-0 indicates a problem.
    """
    configure_logging(args.log_level)
    crash = CrashEntry(args.input)
    LOG.info("Loaded crash %d", crash.crash_id)
    bucket = None
    try:
        # download the crash
        args.input = str(crash.testcase_path())
        if args.sig is None and crash.bucket is not None:
            bucket = Bucket(crash.bucket)
            args.sig = str(bucket.signature_path())
        if args.tool is None:
            args.tool = crash.tool

        # call grizzly.reduce
        result = ReduceManager.main(args)

        # update quality
        # map Session.EXIT_* -> FuzzManagerReporter.QUAL_*
        # default back to UNREDUCED
        # most errors will not be related to the testcase
        # so they should be retried later
        if args.fuzzmanager:
            quality = {
                Session.EXIT_ERROR: FuzzManagerReporter.QUAL_REDUCER_ERROR,
                Session.EXIT_ABORT: crash.testcase_quality,
                Session.EXIT_SUCCESS: FuzzManagerReporter.QUAL_REDUCED_ORIGINAL,
                Session.EXIT_FAILURE: FuzzManagerReporter.QUAL_NOT_REPRODUCIBLE,
            }.get(result, FuzzManagerReporter.QUAL_UNREDUCED)
            # don't ever set things back to Q4, default to Q5 for that case.
            # Q4 is only used in automation, so ABORT should never happen.
            if quality == FuzzManagerReporter.QUAL_REDUCING:
                quality = FuzzManagerReporter.QUAL_UNREDUCED
            LOG.info(
                "reducer finished -> exit(%d) -> Q%d",
                result,
                quality,
            )
            crash.testcase_quality = quality
    finally:
        crash.cleanup()
        if bucket is not None:
            bucket.cleanup()
    return result


if __name__ == "__main__":
    raise SystemExit(main(ReduceFuzzManagerIDArgs().parse_args()))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access

from .target import Target


class SimpleTarget(Target):
    def cleanup(self):
        pass

    def close(self):
        pass

    @property
    def closed(self):
        pass

    def create_report(self, is_hang=False):
        pass

    def detect_failure(self, ignored):
        pass

    def handle_hang(self, ignore_idle=True):
        pass

    def launch(self, _location, _env_mod=None):
        pass

    @property
    def monitor(self):
        return self._monitor

    @property
    def prefs(self):
        pass

    def save_logs(self, *_args, **_kwargs):
        pass


def test_target_01(tmp_path):
    """test creating a simple Target"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    target = SimpleTarget(str(fake_file), str(fake_file), 321, 2, 3)
    assert target.binary == str(fake_file)
    assert target.extension == str(fake_file)
    assert not target.is_idle(0)
    assert target.launch_timeout == 321
    assert target.log_size() == 0
    assert target.log_limit == 2
    assert target.memory_limit == 3
    assert target.monitor is None
    # test stubs
    target.add_abort_token("none!")
    target.dump_coverage()
    target.reverse(1, 2)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Keep statistics about reduction."""
import json
import re
from collections import namedtuple
from copy import deepcopy
from functools import partial
from itertools import zip_longest
from time import time

ReductionStat = namedtuple(
    "ReductionStat", "name, duration, successes, attempts, size, iterations"
)


class _FormatTable:
    """Format tabular data in a table."""

    def __init__(self, columns, formatters, vsep=" | ", hsep="-"):
        """Initialize a FormatTable instance.

        Arguments:
            columns (iterable(str)): List of column names for the table header.
            formatters (iterable(callable)): List of format functions for each column.
                                             None will result in hiding that column.
            vsep (str): Vertical separation between columns.
            hsep (str): Horizontal separation between header and data.
        """
        assert len(columns) == len(formatters)
        self._columns = tuple(
            column for (column, fmt) in zip(columns, formatters) if fmt is not None
        )
        self._formatters = formatters
        self._vsep = vsep
        self._hsep = hsep

    def format_rows(self, rows):
        """Format rows as a table and return a line generator.

        Arguments:
            rows (list(list(str))): Tabular data. Each row must be the same length as
                                    `columns` passed to `__init__`.

        Yields:
            str: Each line of formatted tabular data.
        """
        max_width = [len(col) for col in self._columns]
        formatted = []
        for row in rows:
            assert len(row) == len(self._formatters)
            formatted.append([])
            offset = 0
            for idx, (data, formatter) in enumerate(zip(row, self._formatters)):
                if formatter is None:
                    offset += 1
                    continue
                data = formatter(data)
                max_width[idx - offset] = max(max_width[idx - offset], len(data))
                formatted[-1].append(data)

        # build a format_str to space out the columns with separators using `max_width`
        # the first column is left-aligned, and other fields are right-aligned.
        format_str = self._vsep.join(
            field % (width,)
            for field, width in zip_longest(["%%-%ds"], max_width, fillvalue="%%%ds")
        )
        yield format_str % self._columns
        yield self._hsep * (len(self._vsep) * (len(self._columns) - 1) + sum(max_width))
        for row in formatted:
            yield format_str % tuple(row)


def format_seconds(duration):
    # format H:M:S, and then remove all leading zeros with regex
    minutes, seconds = divmod(int(duration), 60)
    hours, minutes = divmod(minutes, 60)
    result = re.sub("^[0:]*", "", "%d:%02d:%02d" % (hours, minutes, seconds))
    # if the result is all zeroes, ensure one zero is output
    if not result:
        result = "0"
    return result


def _format_duration(duration, total=0):
    result = ""
    if duration is not None:
        if total == 0:
            percent = 0  # pragma: no cover
        else:
            percent = int(100 * duration / total)
        result = format_seconds(duration)
        result += " (%3d%%)" % (percent,)
    return result


def _format_number(number, total=0):
    result = ""
    if number is not None:
        if total == 0:
            percent = 0
        else:
            percent = int(100 * number / total)
        result = "{:n} ({:3d}%)".format(number, percent)
    return result


class ReductionStats:
    """Statistics about reduction"""

    def __init__(self):
        """Initialize a ReductionStats instance."""
        self._data = {"stats": []}

    def __deepcopy__(self, memo):
        """Return a deep copy of this instance."""
        result = type(self)()
        result._data = deepcopy(self._data, memo)  # pylint: disable=protected-access
        return result

    def add(
        self,
        name,
        testcase_size,
        elapsed=None,
        iterations=None,
        attempts=None,
        successes=None,
    ):
        """Record reduction stats for a given point in time:

        - name of the milestone (eg. init, strategy name completed)
        - current testcase size (bytes)
        - elapsed time (seconds)
        - # of iterations
        - # of total attempts
        - # of successful attempts

        Arguments:
            name (str): name of milestone
            testcase_size (int): size of testcase
            elapsed (float or None): seconds elapsed for period recorded
            iterations (int or None): # of iterations performed
            attempts (int or None): # of attempts performed
            successes (int or None): # of attempts successful

        Returns:
            None
        """
        self._data["stats"].append(
            ReductionStat(
                name=name,
                size=testcase_size,
                duration=elapsed,
                iterations=iterations,
                attempts=attempts,
                successes=successes,
            )
        )

    def add_info(self, name, value):
        """Add extra information to be added to the stats report.

        Arguments:
            name (str): key to identify the information
            value (object): Any JSON serializable value.

        Returns:
            None
        """
        assert not self.has_info(name)
        self._data[name] = value

    def has_info(self, name):
        """Check whether the extra information specified by "name"
        already exists in this instance.

        Arguments:
            name (str): key to identify the information

        Returns:
            bool: Whether the named info already exists.
        """
        assert isinstance(name, str)
        return name in self._data

    def add_timed(self, name, testcase_size_cb):
        """Time and record the period leading up to a reduction milestone.
        eg. a strategy being run.

        Arguments:
            name (str): name of milestone
            testcase_size_cb (callable): callback to get testcase size when
                                         context is exited.

        Returns:
            context: Timer context for the period being recorded up to the milestone.

                     Attributes:
                        attempts (int): # of attempts performed during period
                        iterations (int): # of iterations performed during period.
                        successes (int): # of successful attempts during period
        """
        # pylint: disable=no-self-argument
        class _MilestoneTimer:
            def __init__(sub, sub_name, parent=None):
                sub._name = sub_name
                sub._start = None
                sub._attempts = 0
                sub._iterations = 0
                sub._successes = 0
                sub._parent = parent

            def add_attempts(sub, attempts):
                sub._attempts += attempts
                if sub._parent is not None:
                    sub._parent.add_attempts(attempts)

            def add_iterations(sub, iterations):
                sub._iterations += iterations
                if sub._parent is not None:
                    sub._parent.add_iterations(iterations)

            def add_successes(sub, successes):
                sub._successes += successes
                if sub._parent is not None:
                    sub._parent.add_successes(successes)

            def __enter__(sub):
                sub._start = time()
                return sub

            def add_timed(sub, sub_name):
                """Create a new timed period linked to this one."""
                return type(sub)(sub_name, parent=sub)

            def _stop_early(sub, other):
                """Clone and stop the timer. Add as a new stat to `other`.

                Arguments:
                    other (ReductionStats): The stats to add the clone to.

                Returns:
                    None
                """
                elapsed = time() - sub._start
                other.add(
                    name,
                    testcase_size_cb(),
                    elapsed=elapsed,
                    attempts=sub._attempts,
                    successes=sub._successes,
                    iterations=sub._iterations,
                )
                if sub._parent is not None:
                    sub._parent._stop_early(other)  # pylint: disable=protected-access

            def __exit__(sub, exc_type, exc_value, traceback):
                elapsed = time() - sub._start
                self.add(
                    sub._name,
                    testcase_size_cb(),
                    elapsed=elapsed,
                    attempts=sub._attempts,
                    successes=sub._successes,
                    iterations=sub._iterations,
                )

        return _MilestoneTimer(name)

    def copy(self, stop_timer=None):
        """Create a deep copy of this instance.

        Arguments:
            stop_timer (`add_timed` context):
                In-progress `add_timed` call that should be stopped and added to the
                result.

        Returns:
            ReductionStats: Clone of self
        """
        result = deepcopy(self)
        if stop_timer is not None:
            stop_timer._stop_early(result)  # pylint: disable=protected-access
        return result

    def format_lines(self):
        """Format the current stats in a table.

        Yields:
            str: Formatted lines to be output.
        """
        for key, value in self._data.items():
            if key == "stats":
                continue
            yield "%s: %r" % (key, value)
        stats = self._data["stats"]
        tabulator = _FormatTable(
            ReductionStat._fields,
            ReductionStat(
                name=str,
                # duration and attempts are % of total (last), size % of init (first)
                duration=partial(_format_duration, total=stats[-1].duration),
                attempts=partial(_format_number, total=stats[-1].attempts),
                successes=partial(_format_number, total=stats[-1].successes),
                iterations=None,  # hide
                size=partial(_format_number, total=stats[0].size),
            ),
        )
        yield from tabulator.format_rows(stats)

    def add_to_reporter(self, reporter):
        """Add the reducer stats to reported metadata for the given reporter.

        Arguments:
            reporter (FuzzManagerReporter): Reporter to update.

        Returns:
            None
        """
        reporter.add_extra_metadata("reducer-stats", self._data)

    def json(self):
        """Serialize the stats using JSON.

        Returns:
            str: Stats as JSON.
        """
        return json.dumps(self._data, indent=2)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from logging import getLogger
from time import time

from .common import IOManager, Runner, RunResult, Status, TestFile
from .target import TargetLaunchError

__all__ = ("SessionError", "LogOutputLimiter", "Session")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


LOG = getLogger(__name__)


class SessionError(Exception):
    """The base class for exceptions raised by Session"""


class LogOutputLimiter:
    __slots__ = (
        "_delay",
        "_iterations",
        "_launches",
        "_multiplier",
        "_time",
        "_verbose",
    )

    def __init__(self, delay=300, delta_multiplier=2, verbose=False):
        self._delay = delay  # maximum time delay between output
        self._iterations = 1  # next iteration to trigger output
        self._launches = 1  # next launch to trigger output
        self._multiplier = delta_multiplier  # rate to decrease output (iterations)
        self._time = time()
        self._verbose = verbose  # always output

    def ready(self, cur_iter, launches):
        # calculate if a status line should be output
        if self._verbose:
            return True
        ready = False
        if cur_iter >= self._iterations:
            ready = True
            self._iterations *= self._multiplier
        elif launches >= self._launches:
            ready = True
        elif time() - self._delay >= self._time:
            ready = True
        if ready:
            self._time = time()
            self._launches = launches + 1
        return ready


class Session:
    DISPLAY_VERBOSE = 0  # display status every iteration
    DISPLAY_NORMAL = 1  # quickly reduce the amount of output
    EXIT_SUCCESS = 0
    EXIT_ERROR = 1  # unexpected error occurred (invalid input, unhanded exception, etc)
    EXIT_ARGS = 2  # invalid argument
    EXIT_ABORT = 3  # run aborted (ctrl+c, etc)
    EXIT_LAUNCH_FAILURE = 4  # unrelated Target failure (browser startup crash, etc)
    EXIT_FAILURE = 5  # expected results not reproduced (opposite of EXIT_SUCCESS)

    # display warning when target log files exceed limit (25MB)
    TARGET_LOG_SIZE_WARN = 0x1900000

    __slots__ = (
        "_coverage",
        "_relaunch",
        "_report_size",
        "adapter",
        "iomanager",
        "reporter",
        "server",
        "status",
        "target",
    )

    def __init__(
        self,
        adapter,
        reporter,
        server,
        target,
        coverage=False,
        enable_profiling=False,
        relaunch=1,
        report_size=1,
    ):
        assert relaunch > 0
        assert report_size > 0
        self._coverage = coverage
        self._relaunch = relaunch
        self.adapter = adapter
        self.iomanager = None
        self.reporter = reporter
        self.server = server
        self.target = target
        self.status = None
        self.iomanager = IOManager(report_size=report_size)
        self.status = Status.start(enable_profiling=enable_profiling)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self):
        if self.iomanager is not None:
            self.iomanager.cleanup()
        if self.status is not None:
            self.status.cleanup()

    def display_status(self, log_limiter):
        if self.adapter.remaining is not None:
            LOG.info(
                "[I%04d-L%02d-R%02d] %s",
                self.status.iteration,
                self.adapter.remaining,
                self.status.results,
                self.status.test_name,
            )
        elif log_limiter.ready(self.status.iteration, self.target.monitor.launches):
            LOG.info("I%04d-R%02d ", self.status.iteration, self.status.results)

    def generate_testcase(self, time_limit):
        LOG.debug("calling iomanager.create_testcase()")
        test = self.iomanager.create_testcase(self.adapter.name, time_limit)
        LOG.debug("calling self.adapter.generate()")
        with self.status.measure("generate"):
            self.adapter.generate(test, self.iomanager.server_map)
        self.status.test_name = test.input_fname
        if self.target.prefs is not None:
            # TODO: this can likely be improved
            test.add_meta(TestFile.from_file(self.target.prefs, "prefs.js"))
        return test

    def run(
        self,
        ignore,
        time_limit,
        input_path=None,
        iteration_limit=0,
        display_mode=DISPLAY_NORMAL,
    ):
        assert time_limit > 0
        assert iteration_limit >= 0

        LOG.debug("calling adapter.setup()")
        self.adapter.setup(input_path, self.iomanager.server_map)
        LOG.debug("configuring harness")
        self.iomanager.harness = self.adapter.get_harness()

        log_limiter = LogOutputLimiter(verbose=display_mode == self.DISPLAY_VERBOSE)
        # limit relaunch to max iterations if needed
        relaunch = min(self._relaunch, iteration_limit) or self._relaunch
        if self.adapter.remaining is not None:
            assert self.adapter.remaining > 0
            relaunch = min(relaunch, self.adapter.remaining)
        runner = Runner(self.server, self.target, relaunch=relaunch)
        startup_error = False
        while True:
            LOG.debug("- iteration %d -", self.status.iteration + 1)
            self.status.report()
            self.status.iteration += 1

            if self.target.closed:
                # (re-)launch target
                self.iomanager.purge()
                self.adapter.pre_launch()
                if self.iomanager.harness is None:
                    # harness is not in use, open the test case
                    location = runner.location("/grz_current_test", self.server.port)
                else:
                    # harness is in use, open it and it will open the test case.
                    location = runner.location(
                        "/grz_harness",
                        self.server.port,
                        close_after=relaunch,
                        time_limit=time_limit,
                    )
                try:
                    with self.status.measure("launch"):
                        runner.launch(location, max_retries=3, retry_delay=0)
                except TargetLaunchError as exc:
                    short_sig = exc.report.crash_info.createShortSignature()
                    LOG.info(
                        "Result: %s (%s:%s)",
                        short_sig,
                        exc.report.major[:8],
                        exc.report.minor[:8],
                    )
                    self.reporter.submit([], exc.report)
                    exc.report.cleanup()
                    self.status.count_result(short_sig)
                    raise TargetLaunchError(str(exc), None) from None

            # create and populate a test case
            current_test = self.generate_testcase(time_limit)
            # display status
            self.display_status(log_limiter=log_limiter)

            # run test case
            with self.status.measure("execute"):
                result = runner.run(
                    ignore,
                    self.iomanager.server_map,
                    current_test,
                    coverage=self._coverage,
                )
            current_test.duration = result.duration
            # adapter callbacks
            if result.timeout:
                current_test.hang = True
                LOG.debug("calling self.adapter.on_timeout()")
                self.adapter.on_timeout(current_test, result.served)
            else:
                LOG.debug("calling self.adapter.on_served()")
                self.adapter.on_served(current_test, result.served)
            if not result.attempted:
                LOG.warning("Test case was not served")
                LOG.debug("ignoring test case since nothing was served")
                if not current_test.contains(current_test.landing_page):
                    raise SessionError("Test case is missing landing page")
                if result.initial:
                    # since this is the first iteration since the Target launched
                    # something is likely wrong with the Target or Adapter
                    if result.status == RunResult.FAILED:
                        LOG.warning("Delayed startup failure detected")
                    else:
                        LOG.warning("Timeout too short? System too busy?")
                    # ignore this if it did not happen early on
                    # to avoid aborting a fuzzing session unnecessarily
                    if self.status.iteration < 100:
                        startup_error = True
            else:
                if self.adapter.IGNORE_UNSERVED:
                    LOG.debug("removing unserved files from the test case")
                    current_test.purge_optional(result.served)
                self.iomanager.commit()
            # process results
            if result.status == RunResult.FAILED:
                LOG.debug("result detected")
                report = self.target.create_report(is_hang=result.timeout)
                if result.timeout:
                    short_sig = "Potential hang detected"
                else:
                    short_sig = report.crash_info.createShortSignature()
                LOG.info(
                    "Result: %s (%s:%s)", short_sig, report.major[:8], report.minor[:8]
                )
                self.reporter.submit(self.iomanager.tests, report)
                report.cleanup()
                self.status.count_result(short_sig)
            elif result.status == RunResult.IGNORED:
                self.status.ignored += 1
                LOG.info("Ignored (%d)", self.status.ignored)

            if startup_error:
                raise SessionError("Please check Adapter and Target")

            if self.adapter.remaining is not None and self.adapter.remaining < 1:
                # all test cases have been replayed
                LOG.info("Replay Complete")
                break

            if iteration_limit and self.status.iteration >= iteration_limit:
                LOG.info("Hit iteration limit (%d)", iteration_limit)
                break

            # warn about large browser logs
            self.status.log_size = self.target.log_size()
            if self.status.log_size > self.TARGET_LOG_SIZE_WARN:
                LOG.warning(
                    "Large browser logs: %dMBs", (self.status.log_size / 0x100000)
                )
